#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2016/12/12 14:47
@File    : mess.py
@contact : mmmaaaggg@163.com
@desc    :
"""
import os
import time
import functools
from datetime import datetime, date, timedelta
import pytz
import numpy as np
from pandas import Timestamp
import re
import pandas as pd
from collections import OrderedDict
import logging
import warnings
from functools import reduce
import xlrd
import math

logger = logging.getLogger(__name__)
STR_FORMAT_DATE = '%Y-%m-%d'
STR_FORMAT_DATETIME = '%Y-%m-%d %H:%M:%S'
STR_FORMAT_DATETIME2 = '%Y-%m-%d %H:%M:%S.%f'
STR_FORMAT_TIME = '%H:%M:%S'
PATTERN_DATE_FORMAT_RESTRICT = re.compile(r"\d{4}(\D)*\d{2}(\D)*\d{2}")
PATTERN_DATE_FORMAT = re.compile(r"\d{4}(\D)*\d{1,2}(\D)*\d{1,2}")


def active_coroutine(func):
    """装饰器：向前执行第一个 yield 表达式，预激活 func"""
    @functools.wraps(func)
    def primer(*arg, **kwargs):
        gen = func(*arg, **kwargs)
        next(gen)
        return gen

    return primer


def floor(x, precision=0):
    """带小数位精度控制的 floor"""
    if precision == 0:
        return math.floor(x)
    else:
        return math.floor(x * (10 ** precision)) / (10 ** precision)


def ceil(x, precision=0):
    """带小数位精度控制的 ceil"""
    if precision == 0:
        return math.ceil(x)
    else:
        return math.ceil(x * (10 ** precision)) / (10 ** precision)


def is_any(iterable, func):
    """
    查找是否存在任何一个为True的结果，否则返回False
    :param iterable:
    :param func:
    :return:
    """
    for x in iterable:
        if func(x):
            return True
    else:
        return False


def is_not_nan_or_none(x):
    """
    判断是否不是 NAN 或 None
    :param x:
    :return:
    """
    return False if x is None else not((isinstance(x, float) and np.isnan(x)) or isinstance(x, pd.tslib.NaTType))


def is_nan_or_none(x):
    """
    判断是否是 NAN 或 None
    :param x:
    :return:
    """
    return True if x is None else (isinstance(x, float) and np.isnan(x)) or isinstance(x, pd.tslib.NaTType)


def try_2_float(data):
    try:
        return None if data is None else float(data)
    except:
        logger.exception('%s 转化失败', data)
        return None


def split_chunk(l: list, n: int):
    """
    将数组按照给定长度进行分割
    :param l:
    :param n:
    :return:
    """
    for i in range(0, len(l), n):
        yield l[i:i + n]


def zip_split(*args, sep=','):
    """
    将多个字符串，按照 sep 分割对齐，形成元祖数组
    :param args: [str1, str2, ...]
    :param sep: 默认 ,
    :return:
    """
    return list(zip(*[arg.split(sep=sep) for arg in args]))


def unzip_join(tuple_list, sep=','):
    return (sep.join(arg) for arg in zip(*tuple_list))


def populate_obj(model_obj, data_dic: dict, attr_list=None, error_if_no_key=False):
    """
    通过 dict 设置模型对应的属性
    :param model_obj:
    :param data_dic:
    :param attr_list:
    :param error_if_no_key:
    :return:
    """
    for name in (attr_list if attr_list is not None else data_dic.keys()):
        if name in data_dic:
            setattr(model_obj, name, data_dic[name])
        elif error_if_no_key:
            raise KeyError("data_dic 缺少 '%s' key 无法设置到 %s" % (name, model_obj.__class__.__name__))
        else:
            warnings.warn("data_dic 缺少 '%s' key 无法设置到 %s" % (name, model_obj.__class__.__name__))


def log_param_when_exception(func):

    @functools.wraps(func)
    def handler(*arg, **kwargs):
        try:
            return func(*arg, **kwargs)
        except Exception as exp:
            msg = '%s(%s, %s)' % (
                func.__name__,
                ', '.join([str(v) for v in arg]),
                ', '.join(
                    ['{key}={value}'.format(key=str(key), value=str(value))
                     for key, value in kwargs.items()]
                )
            )
            logger.exception(msg)
            raise exp from exp

    return handler


def try_n_times(times=3, sleep_time=3, logger: logging.Logger=None,
                exception=Exception, exception_sleep_time=0,
                exception_exclusion_set=None):
    """
    尝试最多 times 次，异常捕获记录后继续尝试
    :param times:
    :param sleep_time:
    :param logger: 如果异常需要 log 记录则传入参数
    :param exception: 可用于捕获指定异常，默认 Exception
    :param exception_sleep_time: 当出现异常情况下，sleep n 秒
    :return:
    """
    last_invoked_time = [None]

    def wrap_func(func):

        @functools.wraps(func)
        def try_it(*arg, **kwargs):
            for n in range(1, times+1):
                if sleep_time > 0 and last_invoked_time[0] is not None\
                        and (time.time() - last_invoked_time[0]) < sleep_time:
                    time.sleep(sleep_time - (time.time() - last_invoked_time[0]))

                try:
                    ret_data = func(*arg, **kwargs)
                except exception as exp:
                    if logger is not None:
                        logger.exception("第 %d 次调用 %s(%s, %s) 出错", n, func.__name__, arg, kwargs)
                    if exception_exclusion_set is not None:
                        for exp_ex in exception_exclusion_set:
                            if isinstance(exp, exp_ex):
                                raise exp from exp
                    if exception_sleep_time is not None and exception_sleep_time > 0:
                        time.sleep(exception_sleep_time)
                    continue
                finally:
                    last_invoked_time[0] = time.time()

                break
            else:
                ret_data = None

            return ret_data

        return try_it

    return wrap_func


def date_2_str(dt, format=STR_FORMAT_DATE):
    """将日期类型转换为字符串"""
    if dt is not None and type(dt) in (date, datetime, Timestamp):
        dt_str = dt.strftime(format)
    else:
        dt_str = dt
    return dt_str


def datetime_2_str(dt, format=STR_FORMAT_DATETIME):
    if dt is not None and type(dt) in (date, datetime, Timestamp):
        dt_str = dt.strftime(format)
        # print(type(dt), '->', dt_str)
    elif isinstance(dt, pd.DatetimeIndex):
        dt_str = [x.strftime(format) for x in dt]
        # print(type(dt), '-->', dt_str)
    else:
        dt_str = dt
        # print(type(dt), '没有转换', dt)
    return dt_str


def time_2_str(t):
    if t is not None and isinstance(t, timedelta):
        seconds = t.seconds
        hours, minutes = divmod(seconds, 3600)
        minutes, secs = divmod(minutes, 60)
        dt_str = "{0:02d}:{1:02d}:{2:02d}".format(hours, minutes, secs)
        # print(type(t), '->', dt_str)
    else:
        dt_str = str(t)
        # print(type(t), '没有转换', t)
    return dt_str


def date_time_2_str(d, t):
    """将日期与时间组合成  '%Y-%m-%d %H:%M:%S' 字符串 """
    return date_2_str(d) +  ' ' + time_2_str(t)


def str_2_datetime(datetime_str, format=STR_FORMAT_DATETIME):
    if datetime_str is not None:
        if type(datetime_str) == str:
            date_ret = datetime.strptime(datetime_str, format)
        elif type(datetime_str) in (Timestamp, datetime):
            date_ret = datetime_str
        else:
            date_ret = datetime_str
    else:
        date_ret = datetime_str
    return date_ret


def str_2_bytes(input_str):
    """
    用于将 str 类型转换为 bytes 类型
    :param input_str:
    :return:
    """
    return input_str.encode(encoding='GBK')


def bytes_2_str(bytes_str):
    """
    用于将bytes 类型转换为 str 类型
    :param bytes_str:
    :return:
    """
    return str(bytes_str, encoding='GBK')


def timedelta_2_str(td):
    """
    用于将 pd.Timedelta 类型转换为 str 类型
    :param td:
    :return:
    """
    if isinstance(td, pd.Timedelta):
        ret = str(td).split()[-1]
    else:
        ret = td
    return ret


def pattern_data_format(data_str):
    """
    识别日期格式（例如：2017-12-23），并将其翻译成 %Y-%m-%d 类似的格式
    :param data_str:
    :return:
    """
    date_str_format = PATTERN_DATE_FORMAT_RESTRICT.sub(r'%Y\1%m\2%d', data_str)
    if date_str_format == data_str:
        date_str_format = PATTERN_DATE_FORMAT.sub(r'%Y\1%m\2%d', data_str)
    return date_str_format


def try_2_date(something):
    """
    兼容各种格式尝试将 未知对象转换为 date 类型，相对比 str_2_date 消耗资源，支持更多的类型检查，字符串格式匹配
    :param something:
    :return:
    """
    if something is None:
        date_ret = something
    else:
        something_type = type(something)
        if something_type in (int, np.int64, np.int32, np.int16, np.int8):
            something = str(something)
            something_type = type(something)
        if type(something) == str:
            date_str_format = pattern_data_format(something)
            date_ret = datetime.strptime(something, date_str_format).date()
        elif type(something) in (Timestamp, datetime):
            date_ret = something.date()
        else:
            date_ret = something
    return date_ret


def pd_timedelta_2_timedelta(value):
    if isinstance(value, pd.Timedelta):
        # print(value, 'parse to timedelta')
        dt_value = timedelta(seconds=value.seconds)
    else:
        dt_value = value
    return dt_value


def get_first(iterable, func):
    for n in iterable:
        if func(n):
            return n
    return None


def get_first_idx(iterable, func):
    for idx, n in enumerate(iterable):
        if func(n):
            return idx
    return None


def get_last(iterable, func):
    count = len(iterable)
    for n in range(count - 1, -1, -1):
        if func(iterable[n]):
            return iterable[n]
    return None


def get_last_idx(iterable, func):
    """
    获取最后一个符合条件数据的数组索引
    :param iterable:
    :param func:
    :return:
    """
    count = len(iterable)
    for n in range(count - 1, -1, -1):
        if func(iterable[n]):
            return n
    return None


def replace_none_2_str(string, replace=''):
    return replace if string is None else string


def str_2_date(date_str, date_str_format=STR_FORMAT_DATE):
    """
    将日期字符串转换成 date 类型对象，如果字符串为 None 则返回None
    :param date_str: 日期字符串
    :param date_str_format: 日期字符串格式
    :return:
    """
    if date_str is not None:
        if type(date_str) == str:
            date_ret = datetime.strptime(date_str, date_str_format).date()
        elif type(date_str) in (Timestamp, datetime):
            date_ret = date_str.date()
        else:
            date_ret = date_str
    else:
        date_ret = date_str
    return date_ret


def date2datetime(dt):
    """
    date 类型转换问 datetime类型
    :param dt:
    :return:
    """
    return datetime(dt.year, dt.month, dt.day)


def clean_datetime_remove_time_data(atime):
    """
    将时间对象的 时、分、秒 全部清零
    :param atime:
    :return:
    """
    return datetime(atime.year, atime.month, atime.day)


def clean_datetime_remove_ms(atime):
    """
    将时间对象的 毫秒 全部清零
    :param atime:
    :return:
    """
    return datetime(atime.year, atime.month, atime.day, atime.hour, atime.minute, atime.second)


def utc2local(utc):
    localtime = datetime.utcfromtimestamp(utc).replace(tzinfo=pytz.utc).astimezone(pytz.timezone('Asia/Shanghai'))
    return localtime


def get_cntr_kind_name(contract_name):
    left_idx = len(contract_name) - 1
    for num_str in '1234567890':
        idx = contract_name.find(num_str, 0, left_idx)
        if idx == -1:
            continue
        if idx < left_idx:
            left_idx = idx
        if left_idx <= 1:
            break
    # print(lidx, contractname[:lidx])
    return contract_name[:left_idx]


CACHE_FOLDER_PATH_DIC = {}


def get_folder_path(target_folder_name=None, create_if_not_found=True):
    """
    获得系统缓存目录路径
    :param target_folder_name: 缓存目录名称
    :param create_if_not_found: 如果不存在则创建一个目录，默认：True
    :return: 缓存路径
    """
    global CACHE_FOLDER_PATH_DIC
    if target_folder_name is None:
        target_folder_name = 'cache'
    if target_folder_name not in CACHE_FOLDER_PATH_DIC:
        cache_folder_path_tmp = None
        logger.debug('查找数据目录path:')
        parent_folder_path = os.path.abspath(os.curdir)
        par_path = parent_folder_path
        while not os.path.ismount(par_path):
            # print 'parent path = %s'%par_path
            dir_list = os.listdir(par_path)
            for dir_name in dir_list:
                # print d # .strip()
                if dir_name == target_folder_name:
                    cache_folder_path_tmp = os.path.join(par_path, dir_name)
                    logger.debug('<%s>', cache_folder_path_tmp)
                    break
            if cache_folder_path_tmp is not None:
                break
            par_path = os.path.abspath(os.path.join(par_path, os.path.pardir))
        if cache_folder_path_tmp is None:
            if create_if_not_found:
                cache_folder_path_tmp = os.path.abspath(os.path.join(parent_folder_path, target_folder_name))
                logger.debug('<%s> 创建缓存目录', cache_folder_path_tmp)
                os.makedirs(cache_folder_path_tmp)
                CACHE_FOLDER_PATH_DIC[target_folder_name] = cache_folder_path_tmp
        else:
            CACHE_FOLDER_PATH_DIC[target_folder_name] = cache_folder_path_tmp
    return CACHE_FOLDER_PATH_DIC.setdefault(target_folder_name, None)


def get_cache_file_path(cache_folder_name, file_name, create_if_not_found=True):
    """
    返回缓存文件的路径
    :param file_name: 缓存文件名称
    :param cache_folder_name: 缓存folder名称
    :param create_if_not_found: 如果不存在则创建一个目录，默认：True
    :return: 缓存文件路径
    """
    cache_folder_path = get_folder_path(cache_folder_name, create_if_not_found)
    return os.path.join(cache_folder_path, file_name)


def get_df_between_date(data_df, date_frm, date_to):
    """
    该函数仅用于 return_risk_analysis 中计算使用
    :param data_df:
    :param date_frm:
    :param date_to:
    :return:
    """
    if date_frm is not None and date_to is not None:
        new_data_df = data_df[(data_df.Date >= date_frm) & (data_df.Date <= date_to)]
    elif date_frm is not None:
        new_data_df = data_df[data_df.Date >= date_frm]
    elif date_to is not None:
        new_data_df = data_df[data_df.Date <= date_to]
    else:
        new_data_df = data_df
    new_data_df = new_data_df.reset_index(drop=True)
    return new_data_df


def _get_df_between_date_by_index(data_df, date_frm, date_to):
    """
    该函数仅用于 return_risk_analysis 中计算使用
    :param data_df:
    :param date_frm:
    :param date_to:
    :return:
    """
    if date_frm is not None and date_to is not None:
        new_data_df = data_df[(data_df.index >= date_frm) & (data_df.index <= date_to)]
    elif date_frm is not None:
        new_data_df = data_df[data_df.index >= date_frm]
    elif date_to is not None:
        new_data_df = data_df[data_df.index <= date_to]
    else:
        new_data_df = data_df
    return new_data_df


def return_risk_analysis_old(nav_df: pd.DataFrame, date_frm=None, date_to=None, freq='weekly', rf=0.02):
    """
    按列统计 rr_df 收益率绩效
    :param nav_df: 收益率DataFrame，index为日期，每一列为一个产品的净值走势
    :param date_frm: 统计日期区间，可以为空
    :param date_to: 统计日期区间，可以为空
    :param freq: None 自动识别, 'daily' 'weekly' 'monthly'
    :param rf: 无风险收益率，默认 0.02
    :return:
    """
    nav_df.index = [try_2_date(idx) for idx in nav_df.index]
    nav_sorted_df = nav_df.sort_index()
    rr_df = (1 + nav_sorted_df.pct_change().fillna(0)).cumprod()
    rr_df.index = [try_2_date(d) for d in rr_df.index]
    # 计算数据实际频率是日频、周频、月頻
    rr_df_len = rr_df.shape[0]
    day_per_data = (rr_df.index[rr_df_len - 1] - rr_df.index[0]).days / rr_df_len
    if day_per_data <= 0.005:
        freq_real = 'minute'
    elif day_per_data <= 0.2:
            freq_real = 'hour'
    elif day_per_data <= 2:
        freq_real = 'daily'
    elif day_per_data <= 10:
        freq_real = 'weekly'
    else:
        freq_real = 'monthly'
    if freq is None:
        freq = freq_real
    elif freq != freq_real:
        warnings_msg = "data freq wrong, expect %s, but %s was detected" % (freq, freq_real)
        # warnings.warn(warnings_msg)
        # logging.warning(warnings_msg)
        raise ValueError(warnings_msg)

    freq_str = ''
    if freq == 'weekly':
        data_count_per_year = 50
        freq_str = '周'
    elif freq == 'monthly':
        data_count_per_year = 12
        freq_str = '月'
    elif freq == 'daily':
        data_count_per_year = 250
        freq_str = '日'
    elif freq == 'hour':
        data_count_per_year = 1250
        freq_str = '时'
    elif freq == 'minute':
        data_count_per_year = 75000
        freq_str = '分'
    else:
        raise ValueError('freq=%s 只接受 daily weekly monthly 三种之一', freq)
    stat_dic_dic = OrderedDict()
    # rr_df.index = [str_2_date(d) for d in rr_df.index]
    rr_uindex_df = rr_df.reset_index()
    col_name_list = list(rr_uindex_df.columns)
    date_col_name = col_name_list[0]
    col_name_list = col_name_list[1:]
    if type(date_frm) is str:
        date_frm = datetime.strptime(date_frm, '%Y-%m-%d').date()
    if type(date_to) is str:
        date_to = datetime.strptime(date_to, '%Y-%m-%d').date()
    for col_name in col_name_list:
        data_df = rr_uindex_df[[date_col_name, col_name]]
        # print(data_df)
        data_df.columns = ['Date', 'Value']
        data_df = get_df_between_date(data_df, date_frm, date_to)
        data_df.Value = data_df.Value / data_df.Value[0]
        data_df['ret'] = data_df.Value.pct_change().fillna(0)
        date_span = data_df.Date[data_df.index[-1]] - data_df.Date[data_df.index[0]]
        date_span_fraction = 365 / date_span.days if date_span.days > 0 else 1
        # basic indicators
        CAGR = data_df.Value[data_df.index[-1]] ** date_span_fraction - 1
        period_rr = data_df.Value[data_df.index[-1]] - 1
        ann_vol = np.std(data_df.ret, ddof=1) * np.sqrt(data_count_per_year)
        down_side_vol = np.std(data_df.ret[data_df.ret < 0], ddof=1) * np.sqrt(data_count_per_year)
        # WeeksNum = data.shape[0]
        profit_loss_ratio = -np.mean(data_df.ret[data_df.ret > 0]) / np.mean(data_df.ret[data_df.ret < 0])
        win_ratio = len(data_df.ret[data_df.ret >= 0]) / len(data_df.ret)
        min_value = min(data_df.Value)
        final_value = data_df.Value[data_df.index[-1]]
        max_ret = max(data_df.ret)
        min_ret = min(data_df.ret)
        # End of basic indicators
        # max dropdown related
        data_df['mdd'] = data_df.Value / data_df.Value.cummax() - 1
        mdd_size = min(data_df.mdd)
        droparray = pd.Series(data_df.index[data_df.mdd == 0])
        if len(droparray) == 1:
            mdd_max_period = len(data_df.mdd)
        else:
            if float(data_df.Value[droparray.tail(1)]) > float(data_df.Value.tail(1)):
                droparray = droparray.append(pd.Series(data_df.index[-1]), ignore_index=True)
            mdd_max_period = max(droparray.diff().dropna()) - 1
        # End of max dropdown related
        # High level indicators
        sharpe_ratio = (CAGR - rf) / ann_vol
        sortino_ratio = (CAGR - rf) / down_side_vol
        calmar_ratio = CAGR / (-mdd_size)
        #  Natural month return
        j = 1
        for i in data_df.index:
            if i == 0:
                month_ret = pd.DataFrame([[data_df.Date[i], data_df.Value[i]]], columns=('Date', 'Value'))
            else:
                if data_df.Date[i].month != data_df.Date[i - 1].month:
                    month_ret.loc[j] = [data_df.Date[i - 1], data_df.Value[i - 1]]
                    j += 1
        month_ret.loc[j] = [data_df.Date[data_df.index[-1]], data_df.Value[data_df.index[-1]]]
        month_ret['ret'] = month_ret.Value.pct_change().fillna(0)
        max_rr_month = max(month_ret.ret)
        min_rr_month = min(month_ret.ret)
        # End of Natural month return
        data_len = data_df.shape[0]
        date_begin = data_df.Date[0]  # .date()
        date_end = data_df.Date[data_len-1]
        stat_dic = OrderedDict([('起始日期', date_begin),
                                ('截止日期', date_end),
                                ('区间收益率', '%.2f%%' % (period_rr * 100)),
                                ('最终净值', '%.4f' % final_value),
                                ('最低净值', '%.4f' % min_value),
                                ('年化收益率', '%.2f%%' % (CAGR * 100)),
                                ('年化波动率', '%.2f%%' % (ann_vol * 100)),
                                ('年化下行波动率', '%.2f%%' % (down_side_vol * 100)),
                                ('最大回撤', '%.2f%%' % (mdd_size * 100)),
                                ('夏普率', '%.2f' % sharpe_ratio),
                                ('索提诺比率', '%.2f' % sortino_ratio),
                                ('卡马比率', '%.2f' % calmar_ratio),
                                ('盈亏比', '%.2f' % profit_loss_ratio),
                                ('胜率', '%.2f' % win_ratio),
                                ('最长不创新高（%s）' % freq_str, mdd_max_period),
                                ('统计周期最大收益', '%.2f%%' % (max_ret * 100)),
                                ('统计周期最大亏损', '%.2f%%' % (min_ret * 100)),
                                ('最大月收益', '%.2f%%' % (max_rr_month * 100)),
                                ('最大月亏损', '%.2f%%' % (min_rr_month * 100))])
        stat_dic_dic[col_name] = stat_dic
    stat_df = pd.DataFrame(stat_dic_dic)
    stat_df = stat_df.ix[list(stat_dic.keys())]
    return stat_df


def calc_performance(nav_df: pd.DataFrame, date_frm=None, date_to=None, freq='weekly', rf=0.02, suffix_name=None):
    """
    按列统计 rr_df 收益率绩效
    :param nav_df: 收益率DataFrame，index为日期，每一列为一个产品的净值走势
    :param date_frm: 统计日期区间，可以为空
    :param date_to: 统计日期区间，可以为空
    :param freq: None 自动识别, 'daily' 'weekly' 'monthly'
    :param rf: 无风险收益率，默认 0.02
    :return:
    """
    nav_sorted_df = nav_df.copy()
    nav_sorted_df.index = [try_2_date(idx) for idx in nav_sorted_df.index]
    nav_sorted_df.sort_index(inplace=True)
    # 计算数据实际频率是日频、周频、月頻
    data_count = nav_sorted_df.shape[0]
    day_per_data = (nav_sorted_df.index[data_count - 1] - nav_sorted_df.index[0]).days / data_count
    if day_per_data <= 0.008:
        freq_real = 'minute'
    elif day_per_data <= 0.2:
        freq_real = 'hour'
    elif day_per_data <= 2:
        freq_real = 'daily'
    elif day_per_data <= 10:
        freq_real = 'weekly'
    else:
        freq_real = 'monthly'
    if freq is None:
        freq = freq_real
    elif freq != freq_real:
        warnings_msg = "data freq wrong, expect %s, but %s was detected" % (freq, freq_real)
        # warnings.warn(warnings_msg)
        # logging.warning(warnings_msg)
        raise ValueError(warnings_msg)

    freq_str = ''
    if freq == 'weekly':
        data_count_per_year = 50
        freq_str = '周'
    elif freq == 'monthly':
        data_count_per_year = 12
        freq_str = '月'
    elif freq == 'daily':
        data_count_per_year = 250
        freq_str = '日'
    elif freq == 'hour':
        data_count_per_year = 1250
        freq_str = '时'
    elif freq == 'minute':
        data_count_per_year = 75000
        freq_str = '分'
    else:
        raise ValueError('freq=%s 只接受 daily weekly monthly 三种之一', freq)
    stat_dic_dic = OrderedDict()
    if type(date_frm) is str:
        date_frm = datetime.strptime(date_frm, '%Y-%m-%d').date()
    if type(date_to) is str:
        date_to = datetime.strptime(date_to, '%Y-%m-%d').date()

    col_name_list = list(nav_sorted_df.columns)
    # date_col_name = col_name_list[0]
    # col_name_list = col_name_list[1:]
    for col_name in col_name_list:
        data_sub_df = nav_sorted_df[[col_name]].dropna()
        if data_sub_df.shape[0] == 0:
            continue
        # rr_df = (1 + data_sub_df.pct_change().fillna(0)).cumprod()
        # rr_df.index = [try_2_date(d) for d in rr_df.index]
        # data_df = rr_df.reset_index()
        # data_df.columns = ['Date', 'Value']
        # 2018-07-01 不再重置索引，index为日期字段
        data_df = _get_df_between_date_by_index(data_sub_df, date_frm, date_to)
        data_df.columns = ['Value']
        rr_df = data_df.Value.pct_change().fillna(0)
        data_df.Value = (1 + rr_df).cumprod()
        data_df['ret'] = rr_df
        date_list = list(data_df.index)
        date_latest = date_list[-1]
        nav_latest = data_df.Value.loc[date_latest]
        # 计算 近7天，近30天，近365天收益率
        date_week_ago = date_latest - timedelta(days=7)
        date_month_ago = date_latest - timedelta(days=30)
        date_year_ago = date_latest - timedelta(days=365)
        date_week_ago = get_last(date_list, lambda x: x <= date_week_ago)
        date_month_ago = get_last(date_list, lambda x: x <= date_month_ago)
        date_year_ago = get_last(date_list, lambda x: x <= date_year_ago)
        rr_week = (nav_latest / data_df.Value.loc[date_week_ago] - 1) if date_week_ago is not None else None
        rr_month = (nav_latest / data_df.Value.loc[date_month_ago] - 1) if date_month_ago is not None else None
        rr_year = (nav_latest / data_df.Value.loc[date_year_ago] - 1) if date_year_ago is not None else None

        # 计算时间跨度
        date_span = date_list[-1] - date_list[0]
        date_span_fraction = 365 / date_span.days if date_span.days > 0 else 1
        # basic indicators
        CAGR = data_df.Value[date_latest] ** date_span_fraction - 1
        # 相当于余额宝倍数
        times_yeb = (CAGR - 1) / 0.03
        rr_tot = data_df.Value[date_latest] - 1
        ann_vol = np.std(data_df.ret, ddof=1) * np.sqrt(data_count_per_year)
        down_side_vol = np.std(data_df.ret[data_df.ret < 0], ddof=1) * np.sqrt(data_count_per_year)
        # WeeksNum = data.shape[0]
        profit_loss_ratio = -np.mean(data_df.ret[data_df.ret > 0]) / np.mean(data_df.ret[data_df.ret < 0])
        win_ratio = len(data_df.ret[data_df.ret >= 0]) / len(data_df.ret)
        min_value = min(data_df.Value)
        final_value = data_df.Value[data_df.index[-1]]
        max_ret = max(data_df.ret)
        min_ret = min(data_df.ret)
        # End of basic indicators
        # max dropdown related
        data_df['mdd'] = data_df.Value / data_df.Value.cummax() - 1
        mdd_size = min(data_df.mdd)
        droparray = pd.Series(data_df.index[data_df.mdd == 0])
        if len(droparray) == 1:
            mdd_max_period = len(data_df.mdd)
        else:
            if float(data_df.Value[droparray.tail(1)]) > float(data_df.Value.tail(1)):
                droparray = droparray.append(pd.Series(data_df.index[-1]))  # , ignore_index=True
            mdd_max_period = max(droparray.diff().dropna()).days - 1
        # End of max dropdown related
        # High level indicators
        sharpe_ratio = (CAGR - rf) / ann_vol
        sortino_ratio = (CAGR - rf) / down_side_vol
        calmar_ratio = CAGR / (-mdd_size)
        #  Natural month return
        j = 1
        for i, (date_4_df_idx, item) in enumerate(data_df.T.items()):
            if i == 0:
                month_ret = pd.DataFrame([[date_4_df_idx, item.Value]], columns=('Date', 'Value'))
            else:
                date_last_4_last = data_df.index[i - 1]
                if date_4_df_idx.month != date_last_4_last.month:
                    month_ret.loc[j] = [date_last_4_last, data_df.Value[date_last_4_last]]
                    j += 1

        month_ret.loc[j] = [date_latest, nav_latest]
        month_ret['ret'] = month_ret.Value.pct_change().fillna(0)
        max_rr_month = max(month_ret.ret)
        min_rr_month = min(month_ret.ret)
        # End of Natural month return
        date_begin = date_list[0]  # .date()
        date_end = date_list[-1]
        stat_dic = OrderedDict([('date_begen', date_begin),
                                ('date_end', date_end),
                                ('rr_tot', rr_tot),
                                ('rr_week', rr_week),
                                ('rr_month', rr_month),
                                ('rr_year', rr_year),
                                ('final_value', final_value),
                                ('min_value', min_value),
                                ('CAGR', CAGR),
                                ('ann_vol', ann_vol),
                                ('down_side_vol', down_side_vol),
                                ('mdd', mdd_size),
                                ('sharpe_ratio', sharpe_ratio),
                                ('sortino_ratio', sortino_ratio),
                                ('calmar_ratio', calmar_ratio),
                                ('profit_loss_ratio', profit_loss_ratio),  # 盈亏比
                                ('win_ratio', '%.2f' % win_ratio),  # 胜率
                                ('mdd_max_period', mdd_max_period),  # 最长不创新高周期数
                                ('freq', freq_str),  # 周期类型
                                ('max_ret', max_ret),  # 统计周期最大收益
                                ('min_ret', min_ret),  # 统计周期最大亏损
                                ('max_rr_month', max_rr_month),  # 最大月收益
                                ('min_rr_month', min_rr_month),  # 最大月亏损
                                ])
        stat_dic_dic[col_name if suffix_name is None else col_name + "_" + suffix_name] = stat_dic

    return stat_dic_dic


def return_risk_analysis(nav_df: pd.DataFrame, date_frm=None, date_to=None, freq='weekly', rf=0.02, suffix_name=None):
    """
    按列统计 rr_df 收益率绩效
    :param nav_df: 收益率DataFrame，index为日期，每一列为一个产品的净值走势
    :param date_frm: 统计日期区间，可以为空
    :param date_to: 统计日期区间，可以为空
    :param freq: None 自动识别, 'daily' 'weekly' 'monthly'
    :param rf: 无风险收益率，默认 0.02
    :return:
    """
    nav_sorted_df = nav_df.copy()
    nav_sorted_df.index = pd.to_datetime([try_2_date(idx) for idx in nav_sorted_df.index])
    nav_sorted_df.sort_index(inplace=True)
    # 计算数据实际频率是日频、周频、月頻
    data_count = nav_sorted_df.shape[0]
    day_per_data = (nav_sorted_df.index[data_count - 1] - nav_sorted_df.index[0]).days / data_count
    if day_per_data <= 0.008:
        freq_real = 'minute'
    elif day_per_data <= 0.2:
            freq_real = 'hour'
    elif day_per_data <= 2:
        freq_real = 'daily'
    elif day_per_data <= 10:
        freq_real = 'weekly'
    else:
        freq_real = 'monthly'
    if freq is None:
        freq = freq_real
    elif freq != freq_real:
        warnings_msg = "data freq wrong, expect %s, but %s was detected" % (freq, freq_real)
        # warnings.warn(warnings_msg)
        # logging.warning(warnings_msg)
        raise ValueError(warnings_msg)

    freq_str = ''
    if freq == 'weekly':
        data_count_per_year = 50
        freq_str = '周'
    elif freq == 'monthly':
        data_count_per_year = 12
        freq_str = '月'
    elif freq == 'daily':
        data_count_per_year = 250
        freq_str = '日'
    elif freq == 'hour':
        data_count_per_year = 1250
        freq_str = '时'
    elif freq == 'minute':
        data_count_per_year = 75000
        freq_str = '分'
    else:
        raise ValueError('freq=%s 只接受 daily weekly monthly 三种之一', freq)
    stat_dic_dic = OrderedDict()
    mon_rr_dic = {}
    if type(date_frm) is str:
        date_frm = datetime.strptime(date_frm, '%Y-%m-%d').date()
    if type(date_to) is str:
        date_to = datetime.strptime(date_to, '%Y-%m-%d').date()

    col_name_list = list(nav_sorted_df.columns)
    # date_col_name = col_name_list[0]
    # col_name_list = col_name_list[1:]
    for col_name in col_name_list:
        data_sub_df = nav_sorted_df[[col_name]].dropna()
        if data_sub_df.shape[0] == 0:
            continue
        rr_df = (1 + data_sub_df.pct_change().fillna(0)).cumprod()
        # rr_df.index = [try_2_date(d) for d in rr_df.index]
        data_df = rr_df.reset_index()
        data_df.columns = ['Date', 'Value']
        data_df = get_df_between_date(data_df, date_frm, date_to)
        data_df.Value = data_df.Value / data_df.Value[0]
        data_df['ret'] = data_df.Value.pct_change().fillna(0)
        date_span = data_df.Date[data_df.index[-1]] - data_df.Date[data_df.index[0]]
        date_span_fraction = 365 / date_span.days if date_span.days > 0 else 1
        # basic indicators
        CAGR = data_df.Value[data_df.index[-1]] ** date_span_fraction - 1
        period_rr = data_df.Value[data_df.index[-1]] - 1
        ann_vol = np.std(data_df.ret, ddof=1) * np.sqrt(data_count_per_year)
        down_side_vol = np.std(data_df.ret[data_df.ret < 0], ddof=1) * np.sqrt(data_count_per_year)
        # WeeksNum = data.shape[0]
        profit_loss_ratio = -np.mean(data_df.ret[data_df.ret > 0]) / np.mean(data_df.ret[data_df.ret < 0])
        win_ratio = len(data_df.ret[data_df.ret >= 0]) / len(data_df.ret)
        min_value = min(data_df.Value)
        final_value = data_df.Value[data_df.index[-1]]
        max_ret = max(data_df.ret)
        min_ret = min(data_df.ret)
        # End of basic indicators
        # max dropdown related
        data_df['mdd'] = data_df.Value / data_df.Value.cummax() - 1
        mdd_size = min(data_df.mdd)
        droparray = pd.Series(data_df.index[data_df.mdd == 0])
        if len(droparray) == 1:
            mdd_max_period = len(data_df.mdd)
        else:
            if float(data_df.Value[droparray.tail(1)]) > float(data_df.Value.tail(1)):
                droparray = droparray.append(pd.Series(data_df.index[-1]))  # , ignore_index=True
            mdd_max_period = max(droparray.diff().dropna()) - 1
        # End of max dropdown related
        # High level indicators
        sharpe_ratio = (CAGR - rf) / ann_vol
        sortino_ratio = (CAGR - rf) / down_side_vol
        calmar_ratio = CAGR / (-mdd_size)
        #  Natural month return
        j = 1
        for i in data_df.index:
            if i == 0:
                month_ret = pd.DataFrame([[data_df.Date[i], data_df.Value[i]]], columns=('Date', 'Value'))
            else:
                if data_df.Date[i].month != data_df.Date[i - 1].month:
                    month_ret.loc[j] = [data_df.Date[i - 1], data_df.Value[i - 1]]
                    j += 1
        month_ret.loc[j] = [data_df.Date[data_df.index[-1]], data_df.Value[data_df.index[-1]]]
        month_ret['ret'] = month_ret.Value.pct_change().fillna(0)
        max_rr_month = max(month_ret.ret)
        min_rr_month = min(month_ret.ret)
        # End of Natural month return
        data_len = data_df.shape[0]
        date_begin = data_df.Date[0]  # .date()
        date_end = data_df.Date[data_len-1]
        stat_dic = OrderedDict([('起始日期', date_begin),
                                ('截止日期', date_end),
                                ('区间收益率', '%.2f%%' % (period_rr * 100)),
                                ('最终净值', '%.4f' % final_value),
                                ('最低净值', '%.4f' % min_value),
                                ('年化收益率', '%.2f%%' % (CAGR * 100)),
                                ('年化波动率', '%.2f%%' % (ann_vol * 100)),
                                ('年化下行波动率', '%.2f%%' % (down_side_vol * 100)),
                                ('最大回撤', '%.2f%%' % (mdd_size * 100)),
                                ('夏普率', '%.2f' % sharpe_ratio),
                                ('索提诺比率', '%.2f' % sortino_ratio),
                                ('卡马比率', '%.2f' % calmar_ratio),
                                ('盈亏比', '%.2f' % profit_loss_ratio),
                                ('胜率', '%.2f' % win_ratio),
                                ('最长不创新高（%s）' % freq_str, mdd_max_period),
                                ('统计周期最大收益', '%.2f%%' % (max_ret * 100)),
                                ('统计周期最大亏损', '%.2f%%' % (min_ret * 100)),
                                ('最大月收益', '%.2f%%' % (max_rr_month * 100)),
                                ('最大月亏损', '%.2f%%' % (min_rr_month * 100))])
        stat_dic_dic[col_name if suffix_name is None else col_name + "_" + suffix_name] = stat_dic

        # 按时间周期进行相关统计
        data_df = data_df.set_index('Date')[['Value']]
        # data_df_g = data_df.groupby(pd.Grouper(freq='M'))
        # TODO: 首月收益未被计算进去，以后再修复
        monthly_rr_df = data_df.resample('M', convention='end').last().pct_change().fillna(0)
        mon_rr_dic[col_name if suffix_name is None else col_name + "_" + suffix_name] = monthly_rr_df

    if len(stat_dic_dic) > 0:
        stat_df = pd.DataFrame(stat_dic_dic)
        stat_df = stat_df.loc[list(stat_dic.keys())]
    else:
        stat_df = None

    return stat_df, mon_rr_dic


class DataFrame(pd.DataFrame):
    def interpolate_inner(self, columns=None, inplace=False):
        if columns is None:
            columns = list(self.columns)
        data = self if inplace else self.copy()
        for col_name in columns:
            index_not_nan = data.index[~np.isnan(data[col_name])]
            if index_not_nan.shape[0] == 0:
                continue
            index_range = (min(index_not_nan), max(index_not_nan))
            # data[col_name][index_range[0]:index_range[1]].interpolate(inplace=True)
            data[col_name][index_range[0]:index_range[1]] = data[col_name][index_range[0]:index_range[1]].interpolate()
        # print(data)
        if ~inplace:
            return data

    def map(self, func):
        row_count, col_count = self.shape
        columns = list(self.columns)
        indexes = list(self.index)
        for col_num in range(col_count):
            col_val = columns[col_num]
            for row_num in range(row_count):
                row_val = indexes[row_num]
                data_val = self.iloc[row_num, col_num]
                self.iloc[row_num, col_num] = func(col_val, row_val, data_val)
        return self


def reduce_list(funx, data_list, initial=None):
    result_list = []

    def reduce_func(x, y):
        # print(x,y)
        result = funx(x, y)
        result_list.append(result)
        return result

    if initial is None:
        reduce(reduce_func, data_list)
    else:
        reduce(reduce_func, data_list, initial)
    return result_list


def _calc_mdd_4_drawback_analysis(pair, y):
    """
    此函数仅供 drawback_analysis 使用
    用于计算最大回撤使用
    :param pair:
    :param y:
    :return:
    """
    max_y_last = pair[0]
    max_y = max_y_last if max_y_last > y else y
    mdd_last = pair[1]
    keep_max = pair[2]
    dd = y / max_y -1
    if keep_max:
        mdd = dd if dd < mdd_last else mdd_last
    else:
        mdd = dd
    return max_y, mdd, keep_max


def drawback_analysis(data_df, keep_max=False):
    """
    计算给定 DataFrame 数据对应的时间序列最大回撤数据
    :param data_df:
    :return:
    """
    if data_df is None or data_df.shape[0] <= 1:
        mdd_df = None
    else:
        mdd_df = data_df.apply(
            lambda xx: [rr[1] for rr in reduce_list(_calc_mdd_4_drawback_analysis, xx, (xx.iloc[0], 0, keep_max))])
    return mdd_df


def return_risk_analysis_by_xls(file_path, date_col=None, nav_col_list=None, encoding=None):
    """
    读xls文件，对每个sheet进行分析，并最终合并绩效分析报告
    回撤分析分别生成文件显示
    :param file_path:
    :return:
    """
    file_path_no_extention, file_extension = os.path.splitext(file_path)
    if file_extension.lower() == '.csv':
        is_csv_file = True
    else:
        is_csv_file = False
    if is_csv_file:
        sheet_names = ['sheet1']
    else:
        workbook = xlrd.open_workbook(file_path)
        sheet_names = workbook.sheet_names()

    sheet_mdd_df_dic = {}
    stat_df = None
    sheet_mon_rr_dic = {}
    for sheet_name in sheet_names:
        # if sheet_name not in col_names:
        #     continue
        try:
            index_col = 0
            if isinstance(date_col, str):
                if is_csv_file:
                    raise ValueError('csv 文件不支持 index_col 参数为字符串')
                sheet = workbook.sheet_by_name(sheet_name)
                # 取得日期索引后退出
                col_name = sheet.cell_value(0, index_col)
                while not(col_name is None or col_name == ""):
                    if col_name == date_col:
                        break
                    index_col += 1
                    col_name = sheet.cell_value(0, index_col)
            elif isinstance(date_col, int):
                index_col = date_col
            else:
                index_col = 0
            # 默认第0列为日期
            # sheetname Deprecated since version 0.21.0: Use sheet_name instead
            if is_csv_file:
                data_df = pd.read_csv(file_path, index_col=index_col, encoding=encoding)  # 某些版本使用 sheet_name
            else:
                data_df = pd.read_excel(file_path, index_col=index_col, sheet_name=sheet_name)  # 某些版本使用 sheet_name

            if data_df is None or data_df.shape[0] == 0:
                continue
            if nav_col_list is not None:
                data_df = data_df[nav_col_list]
            # 是否带suffix
            if re.search("[S|s]heet", sheet_name) is None:
                suffix_name = sheet_name
            else:
                suffix_name = None
            stat_df_tmp, mon_rr_dic = return_risk_analysis(data_df, freq=None, suffix_name=suffix_name)  # , freq='daily'
            if stat_df is None:
                stat_df = stat_df_tmp
            else:
                stat_df = stat_df.merge(stat_df_tmp, how='outer', left_index=True, right_index=True)

            mdd_df = drawback_analysis(data_df)
            sheet_mdd_df_dic[sheet_name] = mdd_df
            sheet_mon_rr_dic[sheet_name] = mon_rr_dic
        except:
            logging.exception('处理 %s 时失败', sheet_name)
            continue
    return stat_df, sheet_mdd_df_dic, sheet_mon_rr_dic


def merge_nav(df_list, date_from=None):
    """
    合并 df_list 将净值进行合并
    :param df_list:
    :param date_from:
    :return:
    """
    nav_df = None
    for nav_tmp_df in df_list:
        if nav_df is None:
            nav_df = nav_tmp_df
        else:
            nav_df = nav_df.merge(nav_tmp_df, how='outer', right_index=True, left_index=True)
    # 净值拟合
    # def calc_mean(nav_s):
    #     nav_sub_s = nav_s.dropna()
    #     if nav_sub_s.shape[0] == 0:
    #         mean_val = np.nan
    #     else:
    #         mean_val = nav_sub_s.mean()
    #     return mean_val

    pct_df = nav_df.pct_change()
    pct_mean_s = pct_df.mean(axis=1).fillna(0) + 1
    # 进行日期过滤
    if date_from is not None:
        pct_mean_s = pct_mean_s[pct_mean_s.index >= str_2_date(date_from)]
    nav_merged_df = pd.DataFrame({"nav": pct_mean_s.cumprod()})
    stat_df, _ = return_risk_analysis(nav_merged_df, freq=None)
    stat_funds_df, _ = return_risk_analysis(nav_df, freq=None)
    stat_all_df = stat_df.merge(stat_funds_df, how='outer', right_index=True, left_index=True)
    return nav_merged_df, nav_df, stat_all_df


def merge_nav_from_file(file_list, date_from=None):
    """
    从excel或csv文件中读取历史净值数据，进行合并
    :param file_list:
    :param date_from:
    :return:
    """
    df_list = []
    error_dic = {}
    for file_info_dic in file_list:
        # 读取文件
        file_path = file_info_dic['file_path']
        file_path_no_extention, file_extension = os.path.splitext(file_path)
        try:
            if file_extension == '.csv':
                data_df = pd.read_csv(file_path)
            elif file_extension in ('.xls', '.xlsx'):
                data_df = pd.read_excel(file_path, index_col=0).reset_index()
            else:
                error_dic['file type'] = '不支持 %s 净值文件类型' % file_extension
        except:
            error_dic['file read'] = '文件内容读取失败'
            logging.exception('文件内容读取失败：%s', file_path_no_extention)
            continue
        # 设置索引
        if 'date_colum_name' in file_info_dic:
            date_colum_name = file_info_dic['date_colum_name']
            data_df.set_index(date_colum_name, inplace=True)
        else:
            date_colum_name = data_df.columns[0]
            data_df.set_index(date_colum_name, inplace=True)
        # 设置索引日期格式
        data_df.index = [try_2_date(x) for x in data_df.index]
        # 取nav数据
        if 'nav_colum_name_list' in file_info_dic:
            nav_colum_name_list = file_info_dic['nav_colum_name_list']
            if isinstance(nav_colum_name_list, list):
                nav_colum_name_dic = OrderedDict()
                for nav_column_name in nav_colum_name_list:
                    if isinstance(nav_column_name, str):
                        nav_colum_name_dic[nav_column_name] = nav_column_name
                    elif isinstance(nav_column_name, tuple):
                        nav_colum_name_dic[nav_column_name[0]] = nav_column_name[1]
                    else:
                        raise ValueError("%s 列名称无效" % nav_column_name)
                nav_df = data_df[list(nav_colum_name_dic.keys())].rename(columns=nav_colum_name_dic)
            else:
                nav_df = data_df[[nav_colum_name_list]]
        else:
            nav_df = data_df
        # 添加 df_list
        df_list.append(nav_df)
    # 合并
    nav_merged_df, nav_df, stat_df = merge_nav(df_list, date_from)
    return nav_merged_df, nav_df, stat_df


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s: %(levelname)s [%(name)s] %(message)s')
    # # 基金绩效分析
    # from pandas.io.formats.excel import ExcelCell
    # file_path = r'd:\WSPych\fof_ams\Stage\periodic_task\analysis_cache\2016-6-1_2018-6-1\各策略指数走势_按机构.csv'
    # file_path_no_extention, _ = os.path.splitext(file_path)
    # stat_df, sheet_mdd_df_dic, sheet_mon_rr_dic = return_risk_analysis_by_xls(file_path, encoding='GBK')  # , date_col="日期", nav_col_list=['产品净值']
    # if stat_df is not None:
    #     stat_df.to_csv('%s_绩效统计.csv' % file_path_no_extention, encoding='GBK')
    # for sheet_name, mdd_df in sheet_mdd_df_dic.items():
    #     mdd_df.to_csv('%s_%s_最大回撤.csv' % (file_path_no_extention, sheet_name), encoding='GBK')
    # if len(sheet_mon_rr_dic) > 0:
    #     xls_file_path = '%s_%s_月度收益.xls' % (file_path_no_extention, sheet_name)
    #     writer = pd.ExcelWriter(xls_file_path)
    #     try:
    #         for sheet_name, mon_rr_dic in sheet_mon_rr_dic.items():
    #             start_row = 1
    #             for name, monthly_rr_df in mon_rr_dic.items():
    #                 year_set = {trade_date.year for trade_date in monthly_rr_df.index}
    #                 monthly_rr_matrix_df = pd.DataFrame(index=year_set, columns=range(1, 13))
    #                 for trade_date, rr_s in monthly_rr_df.T.items():
    #                     monthly_rr_matrix_df.loc[trade_date.year, trade_date.month] = '%2.2f%%' % (rr_s[0] * 100)
    #                 # 写 excel
    #                 # sheet.write(start_row, 0, name)
    #                 writer.write_cells([ExcelCell(0, 0, name)], sheet_name, startrow=start_row - 1)
    #                 monthly_rr_matrix_df.to_excel(writer, sheet_name, startrow=start_row)
    #                 start_row += len(year_set) + 3
    #     finally:
    #         writer.close()

    # 基金净值合并
    # file_list = [
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\新萌\复华1号历史净值180105(1).xls"},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\新萌\历史净值171017.xls",
    #      'date_colum_name': '净值日期', 'nav_colum_name_list': ['最新净值']},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\新萌\新萌拟合后净值.xlsx",
    #      'date_colum_name': '日期', 'nav_colum_name_list': ['拟合后净值']},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\新萌\投放产品历史净值.csv",
    #      'date_colum_name': 'nav_date', 'nav_colum_name_list': ['nav_acc']},
    # ]

    # file_list = [
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\诚盛投资\诚盛2期Z期净值20171229nav.xlsx",
    #      'date_colum_name': '估值基准', 'nav_colum_name_list': [('单位净值',"诚盛2期Z期净值")]},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\诚盛投资\诚盛1期净值表.xlsx",
    #      'date_colum_name': '日期', 'nav_colum_name_list': ['诚盛1期净值']},
    # ]

    # file_list = [
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\合晟\合晟产品历史净值.csv"},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\展弘\展弘投放产品历史净值.xlsx"},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\新萌\新萌合并业绩后净值.xlsx"},
    #     {'file_path': r"d:\Works\F复华投资\L路演、访谈、评估报告\思勰\思勰合并后净值 2018 - 03 - 09.xls"},
    # ]
    #
    # file_list = [
    #     {'file_path': r'd:\Works\F复华投资\L路演、访谈、评估报告\思勰\思瑞二号周净值.xlsx',
    #      'date_colum_name': 'date', 'nav_colum_name_list': ['累计净值']},
    #     {'file_path': r'd:\Works\F复华投资\L路演、访谈、评估报告\思勰\2016.1-2016.10思勰净值.xlsx'},
    #     {'file_path': r'd:\Works\F复华投资\L路演、访谈、评估报告\思勰\思诚十二号周净值(1).xlsx',
    #      'date_colum_name': 'date', 'nav_colum_name_list': ['累计净值']},
    #     {'file_path': r'd:\Works\F复华投资\L路演、访谈、评估报告\思勰\SM2082-思瑞二号私募投资基金周净值(1).xls',
    #      'date_colum_name': '日期', 'nav_colum_name_list': ['累计净值']},
    # ]

    # nav_merged_df, nav_df, stat_df = merge_nav_from_file(file_list)
    # logging.info("\n%s", nav_merged_df)
    # logging.info("\n%s", nav_df)
    # logging.info("\n%s", stat_df)
    #
    # os.path.dirname(file_list[0]['file_path'])
    # folder_path = os.path.dirname(file_list[0]['file_path'])
    # file_name = "合并后净值.xls"
    # file_path = os.path.join(folder_path, file_name)
    # with pd.ExcelWriter(file_path) as writer:
    #     nav_merged_df.to_excel(writer, sheet_name="合并净值")
    #     nav_df.to_excel(writer, sheet_name="基金净值")
    #     stat_df.to_excel(writer, sheet_name="绩效统计")
    #     writer.save()
    # logging.info("输出文件：\n%s", file_path)

    # 测试 chuck 函数
    # a_list = list(range(1, 17))
    # for b_list in split_chunk(a_list, 4):
    #     print(b_list)
    # for b_list in split_chunk(a_list, 5):
    #     print(b_list)
    # for b_list in split_chunk(a_list, 16):
    #     print(b_list)
    # for b_list in split_chunk(a_list, 17):
    #     print(b_list)

    # 测试 log_param_when_exception 函数
    # @log_param_when_exception
    # def foo(a, b, c=None, *args, **kwargs):
    #     raise Exception('some error')
    #
    # foo(1, 2, 3, 4, e=5, f=6)

    # 测试 try_n_times
    @try_n_times(exception_exclusion_set={ValueError}, logger=logger)
    def foo(n):
        if n == 2:
            raise KeyError("key could't be %d" % n)
        elif n == 3:
            raise ValueError("value could't be %d" % n)
        else:
            return n

    for n in range(5):
        logger.debug('invoke n=%d', n)
        foo(n)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/20 15:12
@File    : trade.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
import logging
from abc import abstractmethod, ABC
from collections import OrderedDict
from datetime import datetime
from functools import partial
from ibats_common.config import config
from ibats_common.backend.orm import OrderDetail, engine_ibats, TradeDetail, PosStatusDetail, TradeAgentStatusDetail
from ibats_common.common import RunMode, ExchangeName, BacktestTradeMode, Action, Direction, PositionDateType
from ibats_common.utils.db import with_db_session

logger = logging.getLogger(__package__)


class TraderAgentBase(ABC):
    """
    交易代理（抽象类），回测交易代理，实盘交易代理的父类
    """

    def __init__(self, stg_run_id, exchange_name, agent_name, **agent_params):
        """
        stg_run_id 作为每一次独立的执行策略过程的唯一标识
        :param stg_run_id:
        :param exchange_name:
        """
        self.stg_run_id = stg_run_id
        self.agent_params = agent_params
        self.logger = logging.getLogger(str(self.__class__))
        self.exchange_name = exchange_name
        self.agent_name = agent_name

    @abstractmethod
    def connect(self):
        raise NotImplementedError()

    @abstractmethod
    def open_long(self, instrument_id, price, vol):
        raise NotImplementedError()

    @abstractmethod
    def close_long(self, instrument_id, price, vol):
        raise NotImplementedError()

    @abstractmethod
    def open_short(self, instrument_id, price, vol):
        raise NotImplementedError()

    @abstractmethod
    def close_short(self, instrument_id, price, vol):
        raise NotImplementedError()

    @abstractmethod
    def get_position(self, instrument_id) -> dict:
        raise NotImplementedError()

    @abstractmethod
    def get_order(self, instrument_id):
        raise NotImplementedError()

    @abstractmethod
    def release(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def datetime_last_update_position(self) -> datetime:
        raise NotImplementedError()

    @property
    @abstractmethod
    def datetime_last_rtn_trade_dic(self) -> dict:
        raise NotImplementedError()

    @property
    @abstractmethod
    def datetime_last_update_position_dic(self) -> dict:
        raise NotImplementedError()

    @property
    @abstractmethod
    def datetime_last_send_order_dic(self) -> dict:
        raise NotImplementedError()

    @property
    @abstractmethod
    def get_balance(self) -> dict:
        raise NotImplementedError()


class BacktestTraderAgentBase(TraderAgentBase):
    """
    供调用模拟交易接口使用
    """

    def __init__(self, stg_run_id, **agent_params):
        super().__init__(stg_run_id, **agent_params)
        # 标示 order 成交模式
        self.trade_mode = agent_params.setdefault('trade_mode', BacktestTradeMode.Order_2_Deal)
        # 账户初始资金
        self.init_cash = agent_params['init_cash']
        # 用来标示当前md，一般执行买卖交易是，对时间，价格等信息进行记录
        self.curr_md_period_type = None
        self.curr_md = None
        # 用来保存历史的 order_detail trade_detail pos_status_detail account_info
        self.order_detail_list = []
        self.trade_detail_list = []
        self.pos_status_detail_dic = OrderedDict()
        self.trade_agent_detail_list = []
        # 持仓信息 初始化持仓状态字典，key为 symbol
        self._pos_status_detail_dic = {}
        self._order_detail_dic = {}
        # 账户信息
        self.trade_agent_status_detail_latest = None
        # 关键 key 信息
        self.timestamp_key = None
        self.symbol_key = None
        self.close_key = None
        # 未成交的订单列表
        self.un_finished_order_list = []

    def set_curr_md(self, period_type, md):
        self.curr_md_period_type = period_type
        self.curr_md = md
        if self.trade_mode != BacktestTradeMode.Order_2_Deal:
            # 根据行情判断是否订单成交
            finished_order_list = []
            for order_detail in self.un_finished_order_list:
                # 如果 开多 或 平空 情况下
                if (order_detail.direction == int(Direction.Long) and order_detail.action == int(Action.Open)) or (
                        order_detail.direction == int(Direction.Short) and order_detail.action != int(Action.Open)
                ):
                    # 目前设置只有价格超越订单价格才能成就，= 的情况不能成就
                    if order_detail.order_price < md[self.close_key]:
                        self._record_trade_detail(order_detail)
                        finished_order_list.append(order_detail)
                else:
                    # 目前设置只有价格跌破订单价格才能成就，= 的情况不能成就
                    if order_detail.order_price > md[self.close_key]:
                        self._record_trade_detail(order_detail)
                        finished_order_list.append(order_detail)

            for order_detail in finished_order_list:
                self.un_finished_order_list.remove(order_detail)

    def set_timestamp_key(self, key):
        self.timestamp_key = key

    def set_symbol_key(self, key):
        self.symbol_key = key

    def set_close_key(self, key):
        self.close_key = key

    def connect(self):
        pass

    def check_key(self):
        if self.timestamp_key is None:
            raise ValueError('timestamp_key 尚未设置，请先调用 set_timestamp_key')
        if self.symbol_key is None:
            raise ValueError('symbol_key 尚未设置，请先调用 set_symbol_key')
        if self.close_key is None:
            raise ValueError('close_key 尚未设置，请先调用 set_close_key')

    @property
    def curr_timestamp(self) -> datetime:
        return self.curr_md[self.timestamp_key]

    @property
    def curr_symbol(self) -> str:
        return self.curr_md[self.symbol_key]

    @property
    def curr_close(self) -> float:
        return float(self.curr_md[self.close_key])

    def _record_order_detail(self, symbol, price: float, vol: int, direction: Direction, action: Action):
        order_date = self.curr_timestamp.date()
        order_detail = OrderDetail(stg_run_id=self.stg_run_id,
                                   trade_agent_key=self.agent_name,
                                   order_date=order_date,
                                   order_time=self.curr_timestamp.time(),
                                   order_millisec=0,
                                   direction=int(direction),
                                   action=int(action),
                                   symbol=symbol,
                                   order_price=float(price),
                                   order_vol=int(vol)
                                   )
        if config.BACKTEST_UPDATE_OR_INSERT_PER_ACTION:
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(order_detail)
                session.commit()
        self.order_detail_list.append(order_detail)
        self._order_detail_dic.setdefault(symbol, []).append(order_detail)
        # 更新成交信息
        # Order_2_Deal 模式：下单即成交
        if self.trade_mode == BacktestTradeMode.Order_2_Deal:
            self._record_trade_detail(order_detail)
        else:
            self.un_finished_order_list.append(order_detail)

    def _record_trade_detail(self, order_detail: OrderDetail):
        """
        根据订单信息保存成交结果
        :param order_detail:
        :return:
        """
        trade_detail = TradeDetail.create_by_order_detail(order_detail)
        self.trade_detail_list.append(trade_detail)
        # 更新持仓信息
        self._record_pos_status_detail(trade_detail)

    def _record_pos_status_detail(self, trade_detail: TradeDetail):
        """
        根据成交信息保存最新持仓信息
        :param trade_detail:
        :return:
        """
        symbol = trade_detail.symbol
        if symbol in self._pos_status_detail_dic:
            pos_status_detail_last = self._pos_status_detail_dic[symbol]
            pos_status_detail = pos_status_detail_last.update_by_trade_detail(trade_detail)
        else:
            pos_status_detail = PosStatusDetail.create_by_trade_detail(trade_detail)
        # 更新
        trade_date, trade_time, trade_millisec = \
            pos_status_detail.trade_date, pos_status_detail.trade_time, pos_status_detail.trade_millisec
        self.pos_status_detail_dic[(trade_date, trade_time, trade_millisec)] = pos_status_detail
        self._pos_status_detail_dic[symbol] = pos_status_detail
        # self.c_save_acount_info(pos_status_detail)

    def _record_trade_agent_status_detail(self) -> TradeAgentStatusDetail:
        stg_run_id, init_cash = self.stg_run_id, self.init_cash
        timestamp_curr = self.curr_timestamp
        trade_date = timestamp_curr.date()
        trade_time = timestamp_curr.time()
        trade_millisec = 0
        # trade_price = float(self.curr_md['close'])
        trade_agent_status_detail = TradeAgentStatusDetail(
            stg_run_id=stg_run_id,
                                                 trade_agent_key=self.agent_name,
                                                 trade_date=trade_date,
                                                 trade_time=trade_time,
                                                 trade_millisec=trade_millisec,
                                                 available_cash=init_cash,
                                                 balance_tot=init_cash,
                                                 )
        if config.BACKTEST_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(trade_agent_status_detail)
                session.commit()
        return trade_agent_status_detail

    def _update_by_pos_status_detail(self) -> TradeAgentStatusDetail:
        """根据 持仓列表更新账户信息"""

        pos_status_detail_dic = self._pos_status_detail_dic
        timestamp_curr = self.curr_timestamp
        trade_agent_status_detail = self.trade_agent_status_detail_latest.create_by_self()
        # 上一次更新日期、时间
        # trade_date_last, trade_time_last, trade_millisec_last = \
        #     trade_agent_status_detail.trade_date, trade_agent_status_detail.trade_time, trade_agent_status_detail.trade_millisec
        # 更新日期、时间
        trade_date = timestamp_curr.date()
        trade_time = timestamp_curr.time()
        trade_millisec = 0

        curr_margin = 0
        close_profit = 0
        position_profit = 0
        floating_pl_chg = 0
        margin_chg = 0
        floating_pl_cum = 0
        for symbol, pos_status_detail in pos_status_detail_dic.items():
            curr_margin += pos_status_detail.margin
            if pos_status_detail.position == 0:
                close_profit += pos_status_detail.floating_pl
            else:
                position_profit += pos_status_detail.floating_pl
            floating_pl_chg += pos_status_detail.floating_pl_chg
            margin_chg += pos_status_detail.margin_chg
            floating_pl_cum += pos_status_detail.floating_pl_cum

        available_cash_chg = floating_pl_chg - margin_chg
        trade_agent_status_detail.curr_margin = curr_margin
        # # 对于同一时间，平仓后又开仓的情况，不能将close_profit重置为0
        # if trade_date == trade_date_last and trade_time == trade_time_last and trade_millisec == trade_millisec_last:
        #     trade_agent_status_detail.close_profit += close_profit
        # else:
        # 一个单位时段只允许一次，不需要考虑上面的情况
        trade_agent_status_detail.close_profit = close_profit

        trade_agent_status_detail.position_profit = position_profit
        trade_agent_status_detail.available_cash += available_cash_chg
        trade_agent_status_detail.floating_pl_cum = floating_pl_cum
        trade_agent_status_detail.balance_tot = trade_agent_status_detail.available_cash + curr_margin

        trade_agent_status_detail.trade_date = trade_date
        trade_agent_status_detail.trade_time = trade_time
        trade_agent_status_detail.trade_millisec = trade_millisec
        if config.BACKTEST_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(trade_agent_status_detail)
                session.commit()
        return trade_agent_status_detail

    def _update_pos_status_detail_by_md(self, pos_status_detail_last) -> PosStatusDetail:
        """创建新的对象，根据 trade_detail 更新相关信息"""
        timestamp_curr = self.curr_timestamp
        trade_date = timestamp_curr.date()
        trade_time = timestamp_curr.time()
        trade_millisec = 0
        trade_price = self.curr_close
        # symbol = md['symbol']

        pos_status_detail = pos_status_detail_last.create_by_self()
        pos_status_detail.cur_price = trade_price
        pos_status_detail.trade_date = trade_date
        pos_status_detail.trade_time = trade_time
        pos_status_detail.trade_millisec = trade_millisec

        # 计算 floating_pl margin
        # instrument_info = config.instrument_info_dic[symbol]
        # multiple = instrument_info['VolumeMultiple']
        # margin_ratio = instrument_info['LongMarginRatio']
        multiple, margin_ratio = 1, 1
        position = pos_status_detail.position
        cur_price = pos_status_detail.cur_price
        avg_price = pos_status_detail.avg_price
        pos_status_detail.margin = position * cur_price * multiple * margin_ratio
        pos_status_detail.margin_chg = pos_status_detail.margin - pos_status_detail_last.margin
        if pos_status_detail.direction == Direction.Long:
            pos_status_detail.floating_pl = (cur_price - avg_price) * position * multiple
        else:
            pos_status_detail.floating_pl = (avg_price - cur_price) * position * multiple
        pos_status_detail.floating_pl_chg = pos_status_detail.floating_pl - pos_status_detail_last.floating_pl
        pos_status_detail.floating_pl_cum += pos_status_detail.floating_pl_chg

        if config.BACKTEST_UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(pos_status_detail)
                session.commit()
        return pos_status_detail

    def update_trade_agent_status_detail(self):
        """
        更新 持仓盈亏数据 汇总统计当前周期账户盈利情况
        :return:
        """
        if self.curr_md is None:
            return
        if self.trade_agent_status_detail_latest is None:
            # self.trade_agent_status_detail_latest = AccountStatusInfo.create(self.stg_run_id, self.init_cash, self.curr_md)
            self.trade_agent_status_detail_latest = self._record_trade_agent_status_detail()
            self.trade_agent_detail_list.append(self.trade_agent_status_detail_latest)

        symbol = self.curr_symbol
        if symbol in self._pos_status_detail_dic:
            pos_status_detail_last = self._pos_status_detail_dic[symbol]
            trade_date = pos_status_detail_last.trade_date
            trade_time = pos_status_detail_last.trade_time
            # 如果当前K线以及更新则不需再次更新。如果当前K线以及有交易产生，则 pos_info 将会在 _save_pos_status_detail 函数中被更新，因此无需再次更新
            if trade_date == self.curr_timestamp.date() and trade_time == self.curr_timestamp.time():
                return
            # 说明上一根K线位置已经平仓，下一根K先位置将记录清除
            if pos_status_detail_last.position == 0:
                del self._pos_status_detail_dic[symbol]
            # 根据 md 数据更新 仓位信息
            # pos_status_detail = pos_status_detail_last.update_by_md(self.curr_md)
            pos_status_detail = self._update_pos_status_detail_by_md(pos_status_detail_last)
            self._pos_status_detail_dic[symbol] = pos_status_detail

        # 统计账户信息，更新账户信息
        # trade_agent_status_detail = self.trade_agent_status_detail_latest.update_by_pos_status_detail(
        #     self._pos_status_detail_dic, self.curr_md)
        trade_agent_status_detail = self._update_by_pos_status_detail()
        self.trade_agent_status_detail_latest = trade_agent_status_detail
        self.trade_agent_detail_list.append(self.trade_agent_status_detail_latest)
        return self.trade_agent_status_detail_latest

    def open_long(self, symbol, price, vol):
        self._record_order_detail(symbol, price, vol, Direction.Long, Action.Open)

    def close_long(self, symbol, price, vol):
        self._record_order_detail(symbol, price, vol, Direction.Long, Action.Close)

    def open_short(self, symbol, price, vol):
        self._record_order_detail(symbol, price, vol, Direction.Short, Action.Open)

    def close_short(self, symbol, price, vol):
        self._record_order_detail(symbol, price, vol, Direction.Short, Action.Close)

    def get_position(self, symbol, **kwargs) -> dict:
        if symbol in self._pos_status_detail_dic:
            pos_status_detail = self._pos_status_detail_dic[symbol]
            position_date_pos_info_dic = {PositionDateType.History: pos_status_detail}
        else:
            position_date_pos_info_dic = None
        return position_date_pos_info_dic

    @property
    def datetime_last_update_position(self) -> datetime:
        return datetime.now()

    @property
    def datetime_last_rtn_trade_dic(self) -> dict:
        raise NotImplementedError()

    @property
    def datetime_last_update_position_dic(self) -> dict:
        raise NotImplementedError()

    @property
    def datetime_last_send_order_dic(self) -> dict:
        raise NotImplementedError()

    def release(self):
        try:
            with with_db_session(engine_ibats) as session:
                session.add_all(self.order_detail_list)
                self.logger.debug("%d 条 order_detail_list 被保存", len(self.order_detail_list))
                session.add_all(self.trade_detail_list)
                self.logger.debug("%d 条 trade_detail_list 被保存", len(self.trade_detail_list))
                session.add_all(self.pos_status_detail_dic.values())
                self.logger.debug("%d 条 pos_status_detail_dic 被保存", len(self.pos_status_detail_dic))
                session.add_all(self.trade_agent_detail_list)
                self.logger.debug("%d 条 trade_agent_detail_list 被保存", len(self.trade_agent_detail_list))
                session.commit()
        except:
            self.logger.exception("release exception")

    def get_order(self, symbol) -> (OrderDetail, None):
        if symbol in self._order_detail_dic:
            return self._order_detail_dic[symbol]
        else:
            return None

    def get_balance(self):
        position_date_pos_info_dic = {key: {PositionDateType.History: pos_status_detail}
                                      for key, pos_status_detail in self._pos_status_detail_dic.items()}
        return position_date_pos_info_dic


trader_agent_class_dic = {
    RunMode.Backtest: {ExchangeName.Default: TraderAgentBase},
    RunMode.Realtime: {ExchangeName.Default: TraderAgentBase}
}


def trader_agent_factory(run_mode: RunMode, stg_run_id, exchange_name: ExchangeName,
                         **trade_agent_params) -> TraderAgentBase:
    """工厂类用来生成相应 TraderAgentBase 实例"""
    trader_agent_class = trader_agent_class_dic[run_mode][exchange_name]
    trader_agent_obj = trader_agent_class(stg_run_id, exchange_name=exchange_name, **trade_agent_params)
    return trader_agent_obj


def register_trader_agent(agent: TraderAgentBase, run_mode: RunMode, exchange_name: ExchangeName = ExchangeName.Default,
                          is_default=True) -> TraderAgentBase:
    """注册 TraderAgent"""
    trader_agent_class_dic[run_mode][exchange_name] = agent
    if is_default:
        trader_agent_class_dic[run_mode][ExchangeName.Default] = agent
    logger.info('注册 %s trade agent[%s] = %s', run_mode, exchange_name, agent)
    return agent


def trader_agent(run_mode: RunMode, exchange_name: ExchangeName = ExchangeName.Default, is_default=True):
    """用来注册 TraderAgent 的装饰器"""
    func = partial(register_trader_agent, run_mode=run_mode, exchange_name=exchange_name, is_default=is_default)
    return func



# -*- coding: utf-8 -*-
"""
Created on 2017/10/9
@author: MG
"""
from datetime import datetime, timedelta
from sqlalchemy import MetaData, Table, Column, Integer, String, DateTime, Boolean, SmallInteger, Date, Time, func
from sqlalchemy.dialects.mysql import DOUBLE, TINYINT
from sqlalchemy.ext.declarative import declarative_base
from pandas import Timedelta
from ibats_common.backend import engines
from ibats_common.utils.db import with_db_session, get_db_session
from ibats_common.common import Action, Direction
from ibats_common.utils.mess import str_2_date, pd_timedelta_2_timedelta, date_2_str, datetime_2_str, date_time_2_str
import logging
from collections import defaultdict
import warnings

engine_ibats = engines.engine_ibats
BaseModel = declarative_base()
# 每一次实务均产生数据库插入或更新动作（默认：否）
UPDATE_OR_INSERT_PER_ACTION = False
_key_idx = defaultdict(lambda: defaultdict(int))


def idx_generator(key1, key2):
    idx = _key_idx[key1][key2]
    idx += 1
    _key_idx[key1][key2] = idx
    return idx


class StgRunInfo(BaseModel):
    """策略运行信息"""

    __tablename__ = 'stg_run_info'
    stg_run_id = Column(Integer, autoincrement=True, primary_key=True)
    stg_name = Column(String(300))
    dt_from = Column(DateTime())
    dt_to = Column(DateTime())
    run_mode = Column(SmallInteger)
    stg_params = Column(String(5000))
    strategy_handler_param = Column(String(1000))
    md_agent_params_list = Column(String(5000))
    trade_agent_params_list = Column(String(5000))


class OrderDetail(BaseModel):
    """订单信息"""

    __tablename__ = 'order_detail'
    stg_run_id = Column(Integer, primary_key=True)
    order_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    order_dt = Column(DateTime)
    order_date = Column(Date)  # 对应行情数据中 ActionDate
    order_time = Column(Time)  # 对应行情数据中 ActionTime
    order_millisec = Column(Integer, default=0, server_default='0')  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)  # -1：空；1：多
    action = Column(Integer)  # 0：关：1：开
    symbol = Column(String(30))
    order_price = Column(DOUBLE)
    order_vol = Column(DOUBLE)  # 订单量

    def __init__(self, stg_run_id=None, trade_agent_key=None, order_dt=None, order_date=None, order_time=None,
                 order_millisec=0, direction=None, action=None, symbol=None, order_price=0, order_vol=0):
        self.stg_run_id = stg_run_id
        self.order_idx = None if stg_run_id is None else idx_generator(stg_run_id, OrderDetail)
        self.trade_agent_key = trade_agent_key
        self.order_dt = date_time_2_str(order_date, order_time) if order_dt is None else order_dt
        self.order_date = order_date
        self.order_time = order_time
        self.order_millisec = order_millisec
        self.direction = direction.value if isinstance(direction, Direction) else direction
        self.action = action.value if isinstance(action, Action) else action
        self.symbol = symbol
        self.order_price = order_price
        self.order_vol = order_vol

    def __repr__(self):
        return f"<OrderDetail(stg_run_id='{self.stg_run_id}' idx='{self.order_idx}', " \
               f"trade_agent_key={self.trade_agent_key}, direction='{Direction(self.direction)}', " \
               f"action='{Action(self.action)}', symbol='{self.symbol}', order_price='{self.order_price}', " \
               f"order_vol='{self.order_vol}')>"

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 order_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM order_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            session.query(OrderDetail).filter(OrderDetail.stg_run_id == stg_run_id).delete()
            session.commit()


class TradeDetail(BaseModel):
    """记录成交信息"""
    __tablename__ = 'trade_detail'
    stg_run_id = Column(Integer, primary_key=True)
    trade_idx = Column(Integer, primary_key=True)  # , comment="成交id"
    trade_agent_key = Column(String(40))
    order_idx = Column(Integer)  # , comment="对应订单id"
    order_price = Column(DOUBLE)  # , comment="原订单价格"
    order_vol = Column(DOUBLE)  # 订单量 , comment="原订单数量"
    trade_dt = Column(DateTime)  # 对应行情数据中 ActionDate
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)  # -1：空；1：多
    action = Column(Integer)  # 0：关：1：开
    symbol = Column(String(30))
    trade_price = Column(DOUBLE)  # , comment="成交价格"
    trade_vol = Column(DOUBLE)  # 订单量 , comment="成交数量"
    margin = Column(DOUBLE, server_default='0')  # 保证金 , comment="占用保证金"
    commission = Column(DOUBLE, server_default='0')  # 佣金、手续费 , comment="佣金、手续费"

    def __init__(self, stg_run_id=None, trade_agent_key=None, order_idx=None, order_price=None, order_vol=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=0, direction=None, action=None,
                 symbol=None, trade_price=None, trade_vol=None, margin=None, commission=None):
        self.stg_run_id = stg_run_id
        self.trade_idx = None if stg_run_id is None else idx_generator(stg_run_id, TradeDetail)
        self.trade_agent_key = trade_agent_key
        self.order_idx = order_idx
        self.order_price = order_price
        self.order_vol = order_vol
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.direction = direction
        self.action = action
        self.symbol = symbol
        self.trade_price = trade_price
        self.trade_vol = trade_vol
        self.margin = margin
        self.commission = commission

    def set_trade_time(self, value):
        if isinstance(value, Timedelta):
            # print(value, 'parse to timedelta')
            self.trade_time = timedelta(seconds=value.seconds)
        else:
            self.trade_time = value

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 trade_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM trade_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            # session.query(PosStatusInfo).filter(PosStatusInfo.stg_run_id.in_([3, 4])).delete(False)  # 仅供实例
            session.query(TradeDetail).filter(TradeDetail.stg_run_id == stg_run_id).delete()
            session.commit()

    @staticmethod
    def create_by_order_detail(order_detail: OrderDetail):
        symbol = order_detail.symbol
        order_price, order_vol = order_detail.order_price, order_detail.order_vol
        # stg_run_id = order_detail.stg_run_id

        # TODO: 增加滑点、成交比例、费率、保证金比例等参数，该参数将和在回测参数中进行设置
        # instrument_info = Config.instrument_info_dic[symbol]
        # multiple = instrument_info['VolumeMultiple']
        # margin_ratio = instrument_info['LongMarginRatio']
        multiple, margin_ratio = 1, 1
        margin = order_vol * order_price * multiple * margin_ratio
        commission = 0
        trade_detail = TradeDetail(stg_run_id=order_detail.stg_run_id,
                                   trade_agent_key=order_detail.trade_agent_key,
                                   order_idx=order_detail.order_idx,
                                   trade_date=order_detail.order_date,
                                   trade_time=order_detail.order_time,
                                   trade_millisec=order_detail.order_millisec,
                                   direction=order_detail.direction,
                                   action=order_detail.action,
                                   symbol=symbol,
                                   order_price=order_price,
                                   order_vol=order_vol,
                                   trade_price=order_price,
                                   trade_vol=order_vol,
                                   margin=margin,
                                   commission=commission
                                   )
        if UPDATE_OR_INSERT_PER_ACTION:
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(trade_detail)
                session.commit()
        return trade_detail


class PosStatusDetail(BaseModel):
    """
    持仓状态数据
    当持仓状态从有仓位到清仓时（position>0 --> position==0），计算清仓前的浮动收益，并设置到 floating_pl 字段最为当前状态的浮动收益
    在调用 create_by_self 时，则需要处理一下，当 position==0 时，floating_pl 直接设置为 0，避免引起后续计算上的混淆
    2018-11-02 当仓位多空切换时（1根K线内多头反手转空头），则浮动收益继续保持之前数字
    """
    __tablename__ = 'pos_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    pos_status_detail_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    trade_idx = Column(Integer)  # , comment="最新的成交id"
    trade_dt = Column(DateTime)  # 每个订单变化生成一条记录 此数据与 AccSumID 对应数据相同
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    direction = Column(TINYINT)
    symbol = Column(String(30))
    position = Column(DOUBLE, default=0.0)
    avg_price = Column(DOUBLE, default=0.0)  # 所持投资品种上一交易日所有交易的加权平均价
    cur_price = Column(DOUBLE, default=0.0)
    floating_pl = Column(DOUBLE, default=0.0)
    floating_pl_chg = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    margin = Column(DOUBLE, default=0.0)
    margin_chg = Column(DOUBLE, default=0.0)
    position_date = Column(Date, default=0)
    logger = logging.getLogger(f'<Table:{__tablename__}')

    def __repr__(self):
        return f"<PosStatusDetail(id='{self.pos_status_detail_idx}', trade_agent_key={self.trade_agent_key}," \
               f" update_dt='{datetime_2_str(self.trade_dt)}', trade_idx='{self.trade_idx}', symbol='{self.symbol}', " \
               f"direction='{self.direction}', position='{self.position}', avg_price='{self.avg_price}')>"

    def __init__(self, stg_run_id=None, trade_agent_key=None, trade_idx=None, trade_dt=None,
                 trade_date=None, trade_time=None, trade_millisec=None, direction=None, symbol=None, position=None,
                 avg_price=None, cur_price=None, floating_pl=0, floating_pl_chg=0, floating_pl_cum=0,
                 margin=0, margin_chg=0, position_date=None):
        self.stg_run_id = stg_run_id
        self.pos_status_detail_idx = None if stg_run_id is None else idx_generator(stg_run_id, PosStatusDetail)
        self.trade_agent_key = trade_agent_key
        self.trade_idx = trade_idx
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.direction = direction
        self.symbol = symbol
        self.position = position
        self.avg_price = avg_price
        self.cur_price = cur_price
        self.floating_pl = floating_pl
        self.floating_pl_chg = floating_pl_chg
        self.floating_pl_cum = floating_pl_cum
        self.margin = margin
        self.margin_chg = margin_chg
        self.position_date = position_date
        # 记录上一状态的浮动收益，用于在计算仓位多空切换时前一状态的浮动收益
        # self.floating_pl_last = floating_pl_last

    @staticmethod
    def create_by_trade_detail(trade_detail: TradeDetail):
        direction, action, instrument_id = trade_detail.direction, trade_detail.action, trade_detail.symbol
        # trade_price, trade_vol, trade_idx = trade_detail.trade_price, trade_detail.trade_vol, trade_detail.trade_idx
        # trade_date, trade_time, trade_millisec = trade_detail.trade_date, trade_detail.trade_time, trade_detail.trade_millisec
        # stg_run_id = trade_detail.stg_run_id
        if action == int(Action.Close):
            raise ValueError('trade_detail.action 不能为 close')
        pos_status_detail = PosStatusDetail(stg_run_id=trade_detail.stg_run_id,
                                            trade_agent_key=trade_detail.trade_agent_key,
                                            trade_idx=trade_detail.trade_idx,
                                            trade_dt=trade_detail.trade_dt,
                                            trade_date=trade_detail.trade_date,
                                            trade_time=trade_detail.trade_time,
                                            trade_millisec=trade_detail.trade_millisec,
                                            direction=trade_detail.direction,
                                            symbol=trade_detail.symbol,
                                            position=trade_detail.trade_vol,
                                            avg_price=trade_detail.trade_price,
                                            cur_price=trade_detail.trade_price,
                                            margin=0,
                                            margin_chg=0,
                                            floating_pl=0,
                                            floating_pl_chg=0,
                                            floating_pl_cum=0,
                                            position_date=trade_detail.trade_date
                                            )
        if UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(pos_status_detail)
                session.commit()
        return pos_status_detail

    def update_by_trade_detail(self, trade_detail: TradeDetail):
        """
        创建新的对象，根据 trade_detail 更新相关信息
        :param trade_detail:
        :return: 
        """
        # 复制前一个持仓状态
        pos_status_detail = self.create_by_self()
        direction, action, symbol = trade_detail.direction, trade_detail.action, trade_detail.symbol
        trade_price, trade_vol, trade_idx = trade_detail.trade_price, trade_detail.trade_vol, trade_detail.trade_idx

        # 获取合约信息
        # instrument_info = Config.instrument_info_dic[symbol]
        # multiple = instrument_info['VolumeMultiple']
        # margin_ratio = instrument_info['LongMarginRatio']
        multiple, margin_ratio = 1, 1

        # 计算仓位、方向、平均价格
        pos_direction_last, position_last, avg_price_last = self.direction, self.position, self.avg_price
        if pos_direction_last == direction:
            if action == Action.Open:
                # 方向相同：开仓：加仓；
                pos_status_detail.avg_price = \
                    (position_last * avg_price_last + trade_price * trade_vol) / (position_last + trade_vol)
                pos_status_detail.position = position_last + trade_vol
            else:
                # 方向相同：关仓：减仓；
                if trade_vol > position_last:
                    raise ValueError("当前持仓%d，平仓%d，错误" % (position_last, trade_vol))
                elif trade_vol == position_last:
                    # 清仓前计算浮动收益
                    # 未清仓的情况将在下面的代码中统一计算浮动收益
                    if pos_status_detail.direction == Direction.Long:
                        pos_status_detail.floating_pl = (trade_price - avg_price_last) * position_last * multiple
                    else:
                        pos_status_detail.floating_pl = (avg_price_last - trade_price) * position_last * multiple

                    pos_status_detail.avg_price = 0
                    pos_status_detail.position = 0

                else:
                    pos_status_detail.avg_price = (position_last * avg_price_last - trade_price * trade_vol) / (
                            position_last - trade_vol)
                    pos_status_detail.position = position_last - trade_vol
        elif position_last == 0:
            pos_status_detail.avg_price = trade_price
            pos_status_detail.position = trade_vol
            pos_status_detail.direction = direction
        else:
            # 方向相反
            raise ValueError("当前仓位：%s %d手，目标操作：%s %d手，请先平仓在开仓" % (
                "多头" if pos_direction_last == Direction.Long else "空头", position_last,
                "多头" if direction == Direction.Long else "空头", trade_vol,
            ))
            # if position == trade_vol:
            #     # 方向相反，量相同：清仓
            #     pos_status_detail.avg_price = 0
            #     pos_status_detail.position = 0
            # else:
            #     holding_amount = position * avg_price
            #     trade_amount = trade_price * trade_vol
            #     position_rest = position - trade_vol
            #     avg_price = (holding_amount - trade_amount) / position_rest
            #     if position > trade_vol:
            #         # 减仓
            #         pos_status_detail.avg_price = avg_price
            #         pos_status_detail.position = position_rest
            #     else:
            #         # 多空反手
            #         self.logger.warning("%s 持%s：%d -> %d 多空反手", self.symbol,
            #                             '多' if direction == int(Direction.Long) else '空', position, position_rest)
            #         pos_status_detail.avg_price = avg_price
            #         pos_status_detail.position = position_rest
            #         pos_status_detail.direction = Direction.Short if direction == int(Direction.Short) else Direction.Long

        # 设置其他属性
        pos_status_detail.cur_price = trade_price
        pos_status_detail.trade_dt = trade_detail.trade_dt
        pos_status_detail.trade_date = trade_detail.trade_date
        pos_status_detail.trade_time = trade_detail.trade_time
        pos_status_detail.trade_millisec = trade_detail.trade_millisec

        # 计算 floating_pl margin
        position_cur = pos_status_detail.position
        # cur_price = pos_status_detail.cur_price
        avg_price_last = pos_status_detail.avg_price
        pos_status_detail.margin = position_cur * trade_price * multiple * margin_ratio
        # 如果当前仓位不为 0 则计算浮动收益
        if position_cur > 0:
            if pos_status_detail.direction == Direction.Long:
                pos_status_detail.floating_pl = (trade_price - avg_price_last) * position_cur * multiple
            else:
                pos_status_detail.floating_pl = (avg_price_last - trade_price) * position_cur * multiple
        # 如果前一状态仓位为 0,  且不是多空切换的情况，则保留上一状态的浮动收益
        if self.position == 0 and self.trade_dt != trade_detail.trade_dt:
            pos_status_detail.margin_chg = pos_status_detail.margin
            pos_status_detail.floating_pl_chg = pos_status_detail.floating_pl
        else:
            pos_status_detail.margin_chg = pos_status_detail.margin - self.margin
            pos_status_detail.floating_pl_chg = pos_status_detail.floating_pl - self.floating_pl

        pos_status_detail.floating_pl_cum += pos_status_detail.floating_pl_chg

        if UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(pos_status_detail)
                session.commit()
        return pos_status_detail

    def create_by_self(self):
        """
        创建新的对象
        若当前对象持仓为0（position==0），则 浮动收益部分设置为0
        :return: 
        """
        position = self.position
        pos_status_detail = PosStatusDetail(stg_run_id=self.stg_run_id,
                                            trade_agent_key=self.trade_agent_key,
                                            trade_idx=self.trade_idx,
                                            trade_dt=self.trade_dt,
                                            trade_date=self.trade_date,
                                            trade_time=self.trade_time,
                                            trade_millisec=self.trade_millisec,
                                            direction=self.direction,
                                            symbol=self.symbol,
                                            position=position,
                                            avg_price=self.avg_price,
                                            cur_price=self.cur_price,
                                            floating_pl=self.floating_pl if position > 0 else 0,
                                            floating_pl_cum=self.floating_pl_cum,
                                            margin=self.margin,
                                            # floating_pl_last=self.floating_pl,
                                            )
        return pos_status_detail

    @staticmethod
    def remove(stg_run_id: int):
        """
        仅作为调试工具使用，删除指定 stg_run_id 相关的 pos_status_detail
        :param stg_run_id: 
        :return: 
        """
        with with_db_session(engine_ibats) as session:
            # session.execute('DELETE FROM pos_status_detail WHERE stg_run_id=:stg_run_id',
            #                 {'stg_run_id': stg_run_id})
            session.query(PosStatusDetail).filter(PosStatusDetail.stg_run_id == stg_run_id).delete()
            session.commit()
        # PosStatusInfo.query.filter


class TradeAgentStatusDetail(BaseModel):
    """持仓状态数据"""
    __tablename__ = 'trade_agent_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    trade_agent_status_detail_idx = Column(Integer, primary_key=True)
    trade_agent_key = Column(String(40))
    trade_dt = Column(DateTime)
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    available_cash = Column(DOUBLE, default=0.0)  # 可用资金, double
    curr_margin = Column(DOUBLE, default=0.0)  # 当前保证金总额, double
    close_profit = Column(DOUBLE, default=0.0)
    position_profit = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    fee_tot = Column(DOUBLE, default=0.0)
    balance_tot = Column(DOUBLE, default=0.0)

    def __init__(self, stg_run_id=None, trade_agent_key=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=None,
                 available_cash=None, curr_margin=None, close_profit=None, position_profit=None, floating_pl_cum=None,
                 fee_tot=None, balance_tot=None):
        self.stg_run_id = stg_run_id
        self.trade_agent_status_detail_idx = None if stg_run_id is None else idx_generator(
            stg_run_id, TradeAgentStatusDetail)
        self.trade_agent_key = trade_agent_key
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.available_cash = available_cash
        self.curr_margin = curr_margin
        self.close_profit = close_profit
        self.position_profit = position_profit
        self.floating_pl_cum = floating_pl_cum
        self.fee_tot = fee_tot
        self.balance_tot = balance_tot

    @staticmethod
    def create(stg_run_id, trade_agent_key, init_cash: int, md: dict):
        """
        根据 md 及 初始化资金 创建对象，默认日期为当前md数据-1天
        :param stg_run_id:
        :param trade_agent_key:
        :param init_cash: 
        :param md: 
        :return: 
        """
        warnings.warn('该函数为范例函数，需要根据实际情况改写', UserWarning)
        trade_date = str_2_date(md['ActionDay']) - timedelta(days=1)
        trade_time = pd_timedelta_2_timedelta(md['ActionTime'])
        trade_millisec = int(md.setdefault('ActionMillisec', 0))
        trade_price = float(md['close'])
        acc_status_detail = TradeAgentStatusDetail(stg_run_id=stg_run_id,
                                                   trade_agent_key=trade_agent_key,
                                                   trade_date=trade_date,
                                                   trade_time=trade_time,
                                                   trade_millisec=trade_millisec,
                                                   available_cash=init_cash,
                                                   balance_tot=init_cash,
                                                   )
        if UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(acc_status_detail)
                session.commit()
        return acc_status_detail

    def create_by_self(self):
        """
        创建新的对象，默认前一日持仓信息的最新价，等于下一交易日的结算价（即 AvePrice）
        :return: 
        """
        trade_agent_status_detail = TradeAgentStatusDetail(stg_run_id=self.stg_run_id,
                                                           trade_agent_key=self.trade_agent_key,
                                                           trade_date=self.trade_date,
                                                           trade_time=self.trade_time,
                                                           trade_millisec=self.trade_millisec,
                                                           available_cash=self.available_cash,
                                                           curr_margin=self.curr_margin,
                                                           close_profit=self.close_profit,
                                                           position_profit=self.position_profit,
                                                           floating_pl_cum=self.floating_pl_cum,
                                                           fee_tot=self.fee_tot,
                                                           balance_tot=self.balance_tot
                                                           )
        return trade_agent_status_detail

    def update_by_pos_status_detail(self, pos_status_detail_dic, md: dict):
        """
        根据 持仓列表更新账户信息
        :param pos_status_detail_dic:
        :return: 
        """
        warnings.warn('该函数为范例函数，需要根据实际情况改写', UserWarning)
        trade_agent_status_detail = self.create_by_self()
        # 更新日期、时间
        trade_date = md['ActionDay']
        trade_time = pd_timedelta_2_timedelta(md['ActionTime'])
        trade_millisec = int(md.setdefault('ActionMillisec', 0))

        available_cash_chg = 0
        curr_margin = 0
        close_profit = 0
        position_profit = 0
        floating_pl_chg = 0
        margin_chg = 0
        floating_pl_cum = 0
        for instrument_id, pos_status_detail in pos_status_detail_dic.items():
            curr_margin += pos_status_detail.margin
            if pos_status_detail.position == 0:
                close_profit += pos_status_detail.floating_pl
            else:
                position_profit += pos_status_detail.floating_pl
            floating_pl_chg += pos_status_detail.floating_pl_chg
            margin_chg += pos_status_detail.margin_chg
            floating_pl_cum += pos_status_detail.floating_pl_cum

        available_cash_chg = floating_pl_chg - margin_chg
        trade_agent_status_detail.curr_margin = curr_margin
        # # 对于同一时间，平仓后又开仓的情况，不能将close_profit重置为0
        # if trade_date == trade_date_last and trade_time == trade_time_last and trade_millisec == trade_millisec_last:
        #     trade_agent_status_detail.close_profit += close_profit
        # else:
        # 一个单位时段只允许一次，不需要考虑上面的情况
        trade_agent_status_detail.close_profit = close_profit

        trade_agent_status_detail.position_profit = position_profit
        trade_agent_status_detail.available_cash += available_cash_chg
        trade_agent_status_detail.floating_pl_cum = floating_pl_cum
        trade_agent_status_detail.balance_tot = trade_agent_status_detail.available_cash + curr_margin

        trade_agent_status_detail.trade_date = trade_date
        trade_agent_status_detail.trade_time = trade_time
        trade_agent_status_detail.trade_millisec = trade_millisec
        if UPDATE_OR_INSERT_PER_ACTION:
            # 更新最新持仓纪录
            with with_db_session(engine_ibats, expire_on_commit=False) as session:
                session.add(trade_agent_status_detail)
                session.commit()
        return trade_agent_status_detail


class StgRunStatusDetail(BaseModel):
    """记录每一单位时刻 所有trade_agent_detail 状态汇总"""
    __tablename__ = 'stg_run_status_detail'
    stg_run_id = Column(Integer, primary_key=True)  # 对应回测了策略 StgRunID 此数据与 AccSumID 对应数据相同
    stg_run_status_detail_idx = Column(Integer, primary_key=True)
    trade_dt = Column(DateTime)
    trade_date = Column(Date)  # 对应行情数据中 ActionDate
    trade_time = Column(Time)  # 对应行情数据中 ActionTime
    trade_millisec = Column(Integer)  # 对应行情数据中 ActionMillisec
    available_cash = Column(DOUBLE, default=0.0)  # 可用资金, double
    curr_margin = Column(DOUBLE, default=0.0)  # 当前保证金总额, double
    close_profit = Column(DOUBLE, default=0.0)
    position_profit = Column(DOUBLE, default=0.0)
    floating_pl_cum = Column(DOUBLE, default=0.0)
    fee_tot = Column(DOUBLE, default=0.0)
    balance_tot = Column(DOUBLE, default=0.0)

    def __init__(self, stg_run_id=None,
                 trade_dt=None, trade_date=None, trade_time=None, trade_millisec=None,
                 available_cash=None, curr_margin=None, close_profit=None, position_profit=None, floating_pl_cum=None,
                 fee_tot=None, balance_tot=None):
        self.stg_run_id = stg_run_id
        self.stg_run_status_detail_idx = None if stg_run_id is None else idx_generator(
            stg_run_id, StgRunStatusDetail)
        self.trade_dt = date_time_2_str(trade_date, trade_time) if trade_dt is None else trade_dt
        self.trade_date = trade_date
        self.trade_time = trade_time
        self.trade_millisec = trade_millisec
        self.available_cash = available_cash
        self.curr_margin = curr_margin
        self.close_profit = close_profit
        self.position_profit = position_profit
        self.floating_pl_cum = floating_pl_cum
        self.fee_tot = fee_tot
        self.balance_tot = balance_tot

    @staticmethod
    def create_by_trade_agent_status_detail_list(stg_run_id, trade_agent_status_detail_list):
        """通过 trade_agent_status_detail_list 更新当前 stg_run_detail 状态"""
        data_len = len(trade_agent_status_detail_list)
        if data_len == 0:
            return None

        stg_run_id = stg_run_id
        trade_dt, trade_date, trade_time, trade_millisec = None, None, None, None
        available_cash, curr_margin, close_profit, position_profit = 0, 0, 0, 0
        floating_pl_cum, fee_tot, balance_tot = 0, 0, 0
        for detail in trade_agent_status_detail_list:
            if trade_dt is None or trade_dt < detail.trade_dt:
                trade_dt = detail.trade_dt
                trade_date = detail.trade_date
                trade_time = detail.trade_time
                trade_millisec = detail.trade_millisec

            available_cash += 0 if detail.available_cash is None else detail.available_cash
            curr_margin += 0 if detail.curr_margin is None else detail.curr_margin
            close_profit += 0 if detail.close_profit is None else detail.close_profit
            position_profit += 0 if detail.position_profit is None else detail.position_profit
            floating_pl_cum += 0 if detail.floating_pl_cum is None else detail.floating_pl_cum
            fee_tot += 0 if detail.fee_tot is None else detail.fee_tot
            balance_tot += 0 if detail.balance_tot is None else detail.balance_tot

        stg_run_status_detail = StgRunStatusDetail(
            stg_run_id=stg_run_id,
            trade_dt=trade_dt,
            trade_date=trade_date,
            trade_time=trade_time,
            trade_millisec=trade_millisec,
            available_cash=available_cash,
            curr_margin=curr_margin,
            close_profit=close_profit,
            position_profit=position_profit,
            floating_pl_cum=floating_pl_cum,
            fee_tot=fee_tot,
            balance_tot=balance_tot,
        )
        return stg_run_status_detail


def init():
    from ibats_common.utils.db import alter_table_2_myisam
    BaseModel.metadata.create_all(engine_ibats)
    alter_table_2_myisam(engine_ibats)
    print("所有表结构建立完成")


if __name__ == "__main__":
    init()
    # 创建user表，继承metadata类
    # Engine使用Schama Type创建一个特定的结构对象
    # stg_info_table = Table("stg_info", metadata, autoload=True)



# -*- coding: utf-8 -*-
"""
Created on 2017/9/2
@author: MG
@contact : mmmaaaggg@163.com
@desc    : 策略基类，所有策略均集成此基类
"""
import pandas as pd
import logging
from collections import defaultdict
from ibats_common.common import PeriodType, ExchangeName, ContextKey

logger_stg_base = logging.getLogger(__name__)


class StgBase:

    def __init__(self):
        self.stg_run_id = None
        # 记录各个md_agent_key、各个周期 md 数据
        self._md_agent_key_period_df_dic = defaultdict(dict)
        # 记录各个md_agent_key、各个周期 md 列信息
        self._md_agent_key_period_df_col_name_list_dic = defaultdict(dict)
        # 记录各个md_agent_key、各个周期 context 信息
        self._md_agent_key_period_context_dic = defaultdict(dict)
        # 记录各个md_agent_key对应的 td_agent_key list
        self._md_td_agent_key_list_map = defaultdict(list)
        # 记录在行情推送过程中最新的一笔md数据
        # self._period_curr_md_dic = {}
        self.trade_agent = None
        self.trade_agent_dic = {}
        self.logger = logging.getLogger(str(self.__class__))
        self._on_period_event_dic = {
            PeriodType.Tick: EventHandlersRelation(PeriodType.Tick,
                                                   self.on_prepare_tick, self.on_tick, pd.DataFrame),
            PeriodType.Min1: EventHandlersRelation(PeriodType.Min1,
                                                   self.on_prepare_min1, self.on_min1, pd.DataFrame),
            PeriodType.Hour1: EventHandlersRelation(PeriodType.Hour1,
                                                    self.on_prepare_hour1, self.on_hour1, pd.DataFrame),
            PeriodType.Day1: EventHandlersRelation(PeriodType.Day1,
                                                   self.on_prepare_day1, self.on_day1, pd.DataFrame),
            PeriodType.Week1: EventHandlersRelation(PeriodType.Week1,
                                                    self.on_prepare_week1, self.on_week1, pd.DataFrame),
            PeriodType.Mon1: EventHandlersRelation(PeriodType.Mon1,
                                                   self.on_prepare_month1, self.on_month1, pd.DataFrame),
        }

    def set_md_td_agent_key_list_map(self, md_td_agent_key_list_map):
        self._md_td_agent_key_list_map = md_td_agent_key_list_map

    def on_timer(self):
        pass

    def load_md_period_df(self, period, md_df, context):
        """初始化加载 md 数据"""
        md_agent_key = context[ContextKey.md_agent_key]
        self._md_agent_key_period_df_dic[md_agent_key][period] = md_df
        self._md_agent_key_period_df_col_name_list_dic[md_agent_key][period] = list(md_df.columns) \
            if isinstance(md_df, pd.DataFrame) else None
        self._md_agent_key_period_context_dic[md_agent_key][period] = context
        # prepare_event_handler = self._on_period_prepare_event_dic[period]
        prepare_event_handler = self._on_period_event_dic[period].prepare_event
        prepare_event_handler(md_df, context)

    def init(self):
        """
        加载历史数据后，启动周期策略执行函数之前
        执行初始化动作，连接 trade_agent
        以后还可以放出其他初始化动作
        :return:
        """
        # self.trade_agent.connect()
        for num, (name, trade_agent) in enumerate(self.trade_agent_dic.items()):
            # ExchangeName.Default 作为默认 trade_agent 在 trade_agent_dic 中存在重复实例，
            # 因此无需对该 key 进行操作
            if name == ExchangeName.Default:
                continue
            self.logger.debug('%d) init trade_agent %s', num, name)
            trade_agent.connect()

    def release(self):
        """
        释放资源
        :return:
        """
        # self.trade_agent.release()
        for num, (name, trade_agent) in enumerate(self.trade_agent_dic.items()):
            # ExchangeName.Default 作为默认 trade_agent 在 trade_agent_dic 中存在重复实例，
            # 因此无需对该 key 进行操作
            if name == ExchangeName.Default:
                continue
            self.logger.debug('%d) release trade_agent %s', num, name)
            trade_agent.release()

    def _on_period_md_append(self, period, md, md_agent_key):
        """
        （仅供 on_period_md_handler 调用使用）
        用于整理接收到的各个周期行情数据
        :param period:
        :param md:
        :return:
        """
        # self._period_curr_md_dic[period] = md
        md_df = pd.DataFrame([md])
        if period in self._md_agent_key_period_df_dic[md_agent_key]:
            col_name_list = self._md_agent_key_period_df_col_name_list_dic[md_agent_key][period]
            md_df_his = self._md_agent_key_period_df_dic[md_agent_key][period].append(md_df[col_name_list])
            self._md_agent_key_period_df_dic[md_agent_key][period] = md_df_his
        else:
            md_df_his = md_df
            self._md_agent_key_period_df_dic[md_agent_key][period] = md_df_his
            self._md_agent_key_period_df_col_name_list_dic[md_agent_key][period] = \
                list(md_df.columns) if isinstance(md_df, pd.DataFrame) else None
            # self.logger.debug('%s -> %s', period, md)
        return md_df_his

    def _on_period_md_event(self, period, md_df_his, md_agent_key):
        """
        （仅供 on_period_md_handler 调用使用）
        用于将各个周期数据传入对应周期事件处理函数
        :param period:
        :param md_df_his:
        :return:
        """
        # event_handler = self._on_period_md_event_dic[period]
        event_handler = self._on_period_event_dic[period].md_event
        context = self._md_agent_key_period_context_dic[md_agent_key][period]
        # self._trade_agent.curr_md = md
        event_handler(md_df_his, context)

    def on_period_md_handler(self, period, md, md_agent_key):
        """响应 period 数据"""
        # 本机测试，延时0.155秒，从分钟K线合成到交易策略端收到数据
        logger_stg_base.debug("%s -> %s", PeriodType(period), md)
        # self._on_period_md_event(period, md_df_his)
        period_event_relation = self._on_period_event_dic[period]
        event_handler = period_event_relation.md_event
        param_type = period_event_relation.param_type
        context = self._md_agent_key_period_context_dic[md_agent_key][period]
        # TODO 由于每一次进入都需要进行判断，增加不必要的计算，考虑通过优化提高运行效率
        if param_type is dict:
            param = md
        elif param_type is pd.DataFrame:
            param = self._on_period_md_append(period, md, md_agent_key)
        else:
            raise ValueError("不支持 %s 类型作为 %s 的事件参数" % (param_type, period))
        event_handler(param, context)

    def on_prepare_tick(self, md_df, context):
        """Tick 历史数据加载执行语句"""
        pass

    def on_prepare_min1(self, md_df, context):
        """1分钟线 历史数据加载执行语句"""
        pass

    def on_prepare_hour1(self, md_df, context):
        """1小时线 历史数据加载执行语句"""
        pass

    def on_prepare_day1(self, md_df, context):
        """1日线 历史数据加载执行语句"""
        pass

    def on_prepare_week1(self, md_df, context):
        """1周线 历史数据加载执行语句"""
        pass

    def on_prepare_month1(self, md_df, context):
        """1月线 历史数据加载执行语句"""
        pass

    def on_tick(self, md_df, context):
        """Tick策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def on_min1(self, md_df, context):
        """1分钟线策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def on_hour1(self, md_df, context):
        """1小时线策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def on_day1(self, md_df, context):
        """1日线策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def on_week1(self, md_df, context):
        """1周线策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def on_month1(self, md_df, context):
        """1月线策略执行语句，需要相应策略实现具体的策略算法"""
        pass

    def open_long(self, instrument_id, price, vol, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.open_long(instrument_id, price, vol)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].open_long(instrument_id, price, vol)

    def close_long(self, instrument_id, price, vol, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.close_long(instrument_id, price, vol)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].close_long(instrument_id, price, vol)

    def open_short(self, instrument_id, price, vol, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.open_short(instrument_id, price, vol)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].open_short(instrument_id, price, vol)

    def close_short(self, instrument_id, price, vol, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.close_short(instrument_id, price, vol)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].close_short(instrument_id, price, vol)

    def get_position(self, instrument_id, trade_agent_key=None, md_agent_key=None, **kwargs) -> dict:
        """
        position_date 作为key， PosStatusInfo 为 val
        返回 position_date_pos_info_dic
        :param instrument_id:
        :param trade_agent_key: trade_agent key值，默认为None，选择默认 trade_agent
        :return:
        """
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.get_position(instrument_id, **kwargs)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].get_position(instrument_id, **kwargs)

    def get_order(self, instrument_id, trade_agent_key=None, md_agent_key=None) -> list:
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.get_order(instrument_id)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].get_order(instrument_id)

    def cancel_order(self, instrument_id, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.cancel_order(instrument_id)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].cancel_order(instrument_id)

    @property
    def datetime_last_update_position(self, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.datetime_last_update_position
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].datetime_last_update_position

    @property
    def datetime_last_rtn_trade_dic(self, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.datetime_last_rtn_trade_dic
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].datetime_last_rtn_trade_dic

    @property
    def datetime_last_update_position_dic(self, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.datetime_last_update_position_dic
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].datetime_last_update_position_dic

    @property
    def datetime_last_send_order_dic(self, trade_agent_key=None, md_agent_key=None):
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.datetime_last_send_order_dic
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].datetime_last_send_order_dic

    def get_balance(self, non_zero_only=True, trade_type_only=True, currency=None, force_refresh=False,
                    trade_agent_key=None, md_agent_key=None) -> dict:
        """
        调用接口 查询 各个币种仓位
        :param non_zero_only: 只保留非零币种
        :param trade_type_only: 只保留 trade 类型币种，frozen 类型的不保存
        :param currency: 只返回制定币种 usdt eth 等
        :param force_refresh: 强制刷新，默认没30秒允许重新查询一次
        :param trade_agent_key: trade_agent key值，默认为None，选择默认 trade_agent
        :return: {'usdt': {<PositionDateType.History: 2>: {'currency': 'usdt', 'type': 'trade', 'balance': 144.09238}}}
        """
        if md_agent_key is None and trade_agent_key is None:
            return self.trade_agent.get_balance(non_zero_only, trade_type_only, currency, force_refresh)
        else:
            if trade_agent_key is None:
                trade_agent_key = self.get_td_agent_key(md_agent_key)
            return self.trade_agent_dic[trade_agent_key].get_balance(
                non_zero_only, trade_type_only, currency, force_refresh)

    def get_holding_currency(self, force_refresh=False, exclude_usdt=True,
                             trade_agent_key=None, md_agent_key=None) -> dict:
        """
        持仓情况dict（非usdt）,仅包含交易状态 type = 'trade' 的记录
        :param force_refresh:
        :param exclude_usdt: 默认为True，剔除 usdt
        :param trade_agent_key: trade_agent key值，默认为None，选择默认 trade_agent
        :param md_agent_key: md_agent_key key值，默认为None，指明与 trade_agent_key 对应的 md_agent_key
        :return:
         {'eos': {<PositionDateType.History: 2>: {'currency': 'eos', 'type': 'trade', 'balance': 144.09238}}}
        """
        cur_balance_dic = self.get_balance(non_zero_only=True, force_refresh=force_refresh,
                                           trade_agent_key=trade_agent_key, md_agent_key=md_agent_key)

        balance_dic = {}
        for currency, dic in cur_balance_dic.items():
            if exclude_usdt and currency == 'usdt':
                continue
            for pos_date_type, dic_sub in dic.items():
                if dic_sub['type'] != 'trade':
                    continue
                balance_dic.setdefault(currency, {})[pos_date_type] = dic_sub
        return balance_dic

    def get_td_agent_keys(self, md_agent_key):
        """返回首个 md_agent_key 对应的 td_agent_key 列表"""
        return self._md_td_agent_key_list_map[md_agent_key]

    def get_td_agent_key(self, md_agent_key):
        """返回首个 md_agent_key 对应的 td_agent_key"""
        key_list = self.get_td_agent_keys(md_agent_key)
        if len(key_list) == 0:
            return None
        else:
            return key_list[0]


class EventHandlersRelation:
    """
    用于记录事件类型与其对应的各种相关事件句柄之间的关系
    """

    def __init__(self, period_type, prepare_event, md_event, param_type):
        self.period_type = period_type
        self.prepare_event = prepare_event
        self.md_event = md_event
        self.param_type = param_type



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/20 15:12
@File    : md.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from datetime import timedelta, datetime
from ibats_common.common import PeriodType, RunMode, ExchangeName
from threading import Thread
import time
import pandas as pd
import logging
from abc import ABC, abstractmethod
from functools import partial
from ibats_common.utils.mess import str_2_date, active_coroutine

logger = logging.getLogger(__package__)


class MdAgentBase(Thread, ABC):

    def __init__(self, instrument_id_list, md_period: PeriodType, exchange_name, agent_name=None,
                 init_load_md_count=None, init_md_date_from=None, init_md_date_to=None, **kwargs):
        if agent_name is None:
            agent_name = f'{exchange_name}.{md_period}'
        self.exchange_name = exchange_name
        super().__init__(name=agent_name, daemon=True)
        self.md_period = md_period
        self.keep_running = None
        self.instrument_id_list = instrument_id_list
        self.init_load_md_count = int(init_load_md_count) if init_load_md_count is not None else None
        self.init_md_date_from = str_2_date(init_md_date_from)
        self.init_md_date_to = str_2_date(init_md_date_to)
        self.logger = logging.getLogger(str(self.__class__))
        self.agent_name = agent_name
        self.params = kwargs
        # 关键 key 信息
        self.timestamp_key = None
        self.symbol_key = None
        self.close_key = None

    def check_key(self):
        """检查 关键 key 信息 是否设置齐全"""
        if self.timestamp_key is None:
            raise ValueError('timestamp_key 尚未设置')
        if self.symbol_key is None:
            raise ValueError('symbol_key 尚未设置')
        if self.close_key is None:
            raise ValueError('close_key 尚未设置')

    @abstractmethod
    def load_history(self, date_from=None, date_to=None, load_md_count=None) -> (pd.DataFrame, dict):
        """
        从mysql中加载历史数据
        实时行情推送时进行合并后供数据分析使用
        :param date_from: None代表沿用类的 init_md_date_from 属性
        :param date_to: None代表沿用类的 init_md_date_from 属性
        :param load_md_count: 0 代表不限制，None代表沿用类的 init_load_md_count 属性，其他数字代表相应的最大加载条数
        :return: md_df 或者
         ret_data {
            'md_df': md_df, 'datetime_key': 'ts_start',
            'date_key': **, 'time_key': **, 'microseconds_key': **
            }
        """

    def load_history_record(self, date_from=None, date_to=None, load_md_count=None):
        """
        加载历史数据，以协程方式逐条推送出去（生成器方法）
        :param date_from: None代表沿用类的 init_md_date_from 属性
        :param date_to: None代表沿用类的 init_md_date_from 属性
        :param load_md_count: 0 代表不限制，None代表沿用类的 init_load_md_count 属性，其他数字代表相应的最大加载条数
        :return: 加载记录数
        """
        his_df_dic = self.load_history(date_from, date_to, load_md_count)
        md_df = his_df_dic['md_df']
        df_len = md_df.shape[0]
        if df_len == 0:
            return df_len

        datetime_key = his_df_dic['datetime_key'] if 'datetime_key' in his_df_dic else None
        date_key = his_df_dic['date_key'] if 'date_key' in his_df_dic else None
        time_key = his_df_dic['time_key'] if 'time_key' in his_df_dic else None
        microseconds_key = his_df_dic['microseconds_key'] if 'microseconds_key' in his_df_dic else None
        # 字段合法性检查
        if datetime_key is None and date_key is None and time_key is None:
            raise KeyError('load_history 方法返回的 key 无效 %s' % his_df_dic.keys())

        num = 0
        for num, md_s in md_df.iterrows():
            if datetime_key is not None:
                datetime_tag = md_s[datetime_key]
            else:
                datetime_tag = datetime.combine(md_s[date_key], md_s[time_key])

            if microseconds_key is not None:
                datetime_tag += timedelta(microseconds=int(md_df[microseconds_key]))
            yield num, datetime_tag, md_s

        return num

    @active_coroutine
    def cor_load_history_record(self, date_from=None, date_to=None, load_md_count=None):
        """
        加载历史数据，以协程方式逐条推送出去（协程方法）
        :param date_from: None代表沿用类的 init_md_date_from 属性
        :param date_to: None代表沿用类的 init_md_date_from 属性
        :param load_md_count: 0 代表不限制，None代表沿用类的 init_load_md_count 属性，其他数字代表相应的最大加载条数
        :return: 加载记录数        """
        data_count = yield from self.load_history_record(date_from, date_to, load_md_count)
        return data_count

    @abstractmethod
    def connect(self):
        """链接redis、初始化历史数据"""

    @abstractmethod
    def release(self):
        """释放channel资源"""

    def subscribe(self, instrument_id_list=None):
        """订阅合约"""
        if instrument_id_list is None:
            return
        self.instrument_id_list = list(set(self.instrument_id_list) | set(instrument_id_list))

    def unsubscribe(self, instrument_id_list):
        """退订合约"""
        if instrument_id_list is None:
            self.instrument_id_list = []
        else:
            self.instrument_id_list = list(set(self.instrument_id_list) - set(instrument_id_list))


md_agent_class_dic = {
    RunMode.Backtest: {ExchangeName.Default: MdAgentBase},
    RunMode.Realtime: {ExchangeName.Default: MdAgentBase}
}


def md_agent_factory(run_mode: RunMode, instrument_id_list: list, md_period: PeriodType, agent_name=None,
                     exchange_name: ExchangeName = ExchangeName.Default, **kwargs) -> MdAgentBase:
    """工厂类用来生成相应 MdAgentBase 实例"""
    md_agent_class = md_agent_class_dic[run_mode][exchange_name]
    md_agent_obj = md_agent_class(
        instrument_id_list=instrument_id_list, md_period=md_period, agent_name=agent_name, exchange_name=exchange_name,
        **kwargs)
    return md_agent_obj


def register_md_agent(agent: MdAgentBase, run_mode: RunMode, exchange_name: ExchangeName = ExchangeName.Default,
                      is_default=True) -> MdAgentBase:
    """注册 MdAgent"""
    md_agent_class_dic[run_mode][exchange_name] = agent
    if is_default:
        md_agent_class_dic[run_mode][ExchangeName.Default] = agent
    logger.info('注册 %s md agent[%s] = %s', run_mode, exchange_name, agent)
    return agent


def md_agent(run_mode: RunMode, exchange_name: ExchangeName = ExchangeName.Default, is_default=True):
    """用来注册 MdAgent 的装饰器"""
    func = partial(register_md_agent, run_mode=run_mode, exchange_name=exchange_name, is_default=is_default)
    return func


def _test_only():
    instrument_id_list = ['jm1711', 'rb1712', 'pb1801', 'IF1710']
    md_agent_obj = md_agent_factory(RunMode.Realtime, instrument_id_list, md_period=PeriodType.Min1,
                                    init_load_md_count=100)
    md_df = md_agent_obj.load_history()
    print(md_df.shape)
    md_agent_obj.connect()
    md_agent_obj.subscribe(instrument_id_list)
    md_agent_obj.start()
    for n in range(120):
        time.sleep(1)

    md_agent_obj.keep_running = False
    md_agent_obj.join()
    md_agent_obj.release()
    print("all finished")


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(levelname)s %(name)s %(filename)s.%(funcName)s:%(lineno)d|%(message)s')
    _test_only()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
@author  : MG
@Time    : 2018/6/20 15:15
@File    : common.py
@contact : mmmaaaggg@163.com
@desc    : 
"""
from enum import IntEnum, unique, Enum


@unique
class PeriodType(IntEnum):
    """
    周期类型
    """
    Tick = 0
    Min1 = 1
    Min5 = 2
    Min10 = 3
    Min15 = 4
    Min30 = 5
    Hour1 = 10
    Day1 = 20
    Week1 = 30
    Mon1 = 40
    Year1 = 100

    @staticmethod
    def get_min_count(period_type) -> int:
        """
        返回给的周期类型对应的分钟数
        :param period_type:
        :return:
        """
        if PeriodType.Min1 == period_type:
            min_count = 1
        elif PeriodType.Min5 == period_type:
            min_count = 5
        elif PeriodType.Min10 == period_type:
            min_count = 10
        elif PeriodType.Min15 == period_type:
            min_count = 15
        elif PeriodType.Min30 == period_type:
            min_count = 30
        else:
            raise ValueError('不支持 %s 周期' % period_type)
        return min_count


@unique
class RunMode(IntEnum):
    """
    运行模式，目前支持两种：实时行情模式，回测模式
    """
    Realtime = 0
    Backtest = 1


@unique
class ExchangeName(IntEnum):
    """
    交易所名称
    """
    Default = 0
    HuoBi = 1
    BitMex = 2
    DataIntegration = 3


@unique
class Direction(IntEnum):
    """买卖方向"""
    Short = -1  # 空头
    Long = 1  # 多头


@unique
class Action(IntEnum):
    """开仓平仓"""
    Open = 0  # 开仓
    Close = 1  # 平仓
    ForceClose = 2  # 强平
    CloseToday = 3  # 平今
    CloseYesterday = 4  # 平昨
    ForceOff = 5  # 强减
    LocalForceClose = 6  # 本地强平

    @staticmethod
    def create_by_offsetflag(offset_flag):
        """
        将 Api 中 OffsetFlag 变为 Action 类型
        :param offset_flag:
        :return:
        """
        if offset_flag == '0':
            return Action.Open
        elif offset_flag in {'1', '5'}:
            return Action.Close
        elif offset_flag == '3':
            return Action.CloseToday
        elif offset_flag == '4':
            return Action.CloseYesterday
        elif offset_flag == '2':
            return Action.ForceClose
        elif offset_flag == '6':
            return Action.LocalForceClose
        else:
            raise ValueError('Action不支持 %s' % offset_flag)


@unique
class PositionDateType(IntEnum):
    """今日持仓历史持仓标示"""
    Today = 1  # 今日持仓
    History = 2  # 历史持仓


@unique
class BacktestTradeMode(IntEnum):
    """
    回测模式下的成交模式
    """
    Order_2_Deal = 0  # 一种简单回测模式，无论开仓平仓等操作，下单即成交
    MD_2_Deal = 1  # 根据下单后行情变化确定何时成交


@unique
class ContextKey(IntEnum):
    """
    策略执行逻辑中 context[key] 的部分定制化key
    """
    instrument_id_list = 0
    md_agent_key = 1
    period = 2


